# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.sketchpad import Sketchpad as SketchpadRuntime
from libqtopensesame.items.qtplugin import QtPlugin
from libqtopensesame.items.feedpad import Feedpad
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'sketchpad', category=u'item')


class Sketchpad(Feedpad, QtPlugin, SketchpadRuntime):

    r"""The sketchpad controls are implemented in feedpad."""
    description = _(u'Displays stimuli')

    def __init__(self, name, experiment, string=None):

        SketchpadRuntime.__init__(self, name, experiment, string)
        QtPlugin.__init__(self)

    def init_edit_widget(self):
        r"""Initializes the widget."""
        Feedpad.init_edit_widget(self)
        self.sketchpad_widget.ui.widget_settings_reset_variables.hide()


# Alias for backwards compatibility
sketchpad = Sketchpad



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from qtpy import QtWidgets
from libopensesame.python_workspace_api import xy_from_polar
from libqtopensesame.dialogs.base_dialog import BaseDialog


class PolarCoordinates(BaseDialog):

    r"""A dialog for determining cartesian coordinates from polar coordinates."""
    def __init__(self, main_window):
        r"""Constructor.

        Parameters
        ----------
        main_window
            The main window object.
        """
        super().__init__(main_window, ui=u'dialogs.polar_coordinates')
        self.update()

    def xy(self):

        return xy_from_polar(
            pole=(self.ui.spinbox_pole_x.value(),
                  self.ui.spinbox_pole_y.value()),
            rho=self.ui.spinbox_rho.value(),
            phi=self.ui.spinbox_phi.value()
        )

    def update(self):

        super().update()
        x, y = self.xy()
        self.ui.label_cartesian_x.setText(u'%.2f px' % x)
        self.ui.label_cartesian_y.setText(u'%.2f px' % y)

    def get_xy(self):
        r"""Executes the dialog and gets the xy coordinates, or None if the
        dialog was cancelled.

        Returns
        -------
        tuple, NoneType
        """
        if self.exec_() != QtWidgets.QDialog.Accepted:
            return None
        return self.xy()


# Alias for backwards compatibility
polar_coordinates = PolarCoordinates



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from openexp._sampler.sampler import Sampler
from libopensesame.oslogging import oslogger
from openexp.backend import configurable
from openexp.keyboard import Keyboard
import sys
import soundfile as sf
import numpy as np

DEFAULT_SOUND_FREQ = 48000
DEFAULT_BLOCK_SIZE = 256
NEEDS_BLOCK_SIZE = 'sounddevice', 'PTB'
# Due to issues with sounddevice/ portaudio on Mac OS, we fall back to
# pygame on that platform
DEFAULT_AUDIOLIB = 'pygame' if sys.platform == 'darwin' else 'sounddevice'
# Will be initialized during init_sound()
Sound = None
PLAYING = None



class Psycho(Sampler):
    """This is a sampler backend built on top of PsychoPy. For function
    specifications and docstrings, see `openexp._sampler.sampler`.
    """
    # The settings variable is used by the GUI to provide a list of back-end
    # settings
    settings = {
        'psycho_audiolib': {
            'name': 'Sound library',
            'description': 'Can be sounddevice, pyo, pygame, or PTB',
            'default': DEFAULT_AUDIOLIB},
        'sound_freq': {
            'name': 'Sampling frequency for synth',
            'description': 'Determines the sampling rate of synthesized sounds',
            'default': DEFAULT_SOUND_FREQ},
        'block_size': {
            'name': 'Buffer size',
            'description': 'Low values for low latency. High values for stability.',
            'default': DEFAULT_BLOCK_SIZE}}

    def __init__(self, experiment, src, **playback_args):

        if (isinstance(src, np.ndarray) and src.dtype == np.int16 and
                src.ndim == 1):
            # The Synth provides the data as a 1D array with int values between
            # 0 and 32767. The signal is stereo but flattened into a single
            # trace, so we expand it here.
            self._data = np.array(src, dtype=np.float64) / 32767
            self._data = self._data.reshape((self._data.shape[0] // 2, 2))
            self._samplerate = experiment.var.get('sound_freq',
                                                  DEFAULT_SOUND_FREQ)
        else:
            self._data, self._samplerate = sf.read(src)
        # Create keyword arguments, which depend on the sound backend
        kwargs = {}
        if (experiment.var.get('psycho_audiolib', DEFAULT_AUDIOLIB)
                in NEEDS_BLOCK_SIZE):
            kwargs['blockSize'] = experiment.var.get('block_size',
                                                     DEFAULT_BLOCK_SIZE)
        # Make sure that the data is a [samples, 2] array for stereo data
        if self._data.ndim == 1:
            self._data.shape = [len(self._data), 1]
        if self._data.shape[1] == 1:
            self._data = self._data.repeat(2, axis=1)
        self._sound = Sound(self._data, sampleRate=self._samplerate, **kwargs)
        # isPlaying() was introduced recently. For older versions of psychopy
        # we monkeypatch this function into existence using the older status
        # property
        # - https://github.com/psychopy/psychopy/commit/\
        #   50a730b7be0bb1a219d5666d657bad4c8cc121d1
        if not hasattr(self._sound, 'isPlaying'):
            self._sound.isPlaying = lambda self: self.status == PLAYING
        self._keyboard = Keyboard(experiment)
        Sampler.__init__(self, experiment, src, **playback_args)

    def set_config(self, **cfg):

        Sampler.set_config(self, **cfg)
        if not cfg:
            return
        if 'volume' in cfg:
            self._set_volume(cfg['volume'])
        if 'pitch' in cfg:
            self._set_pitch(cfg['pitch'])
        if 'pan' in cfg:
            self._set_pan(cfg['pan'])
        if 'duration' in cfg:
            self._set_duration(cfg['duration'])
        if 'fade_in' in cfg:
            self._set_fade_in(cfg['fade_in'])
        self._sound.stopTime = -1  # Fixes a bug in PsychoPy 3.2.4
        self._sound.setSound(self._data)

    def _set_volume(self, volume):

        if volume == 1:
            return
        self._data *= volume

    def _set_pitch(self, pitch):

        from scipy.signal import resample
        if pitch == 1:
            return
        self._data = resample(self._data, int(self._data.shape[0] / pitch))

    def _set_pan(self, pan):

        if pan == 'left':
            right_volume = 0.
            left_volume = 1.
        elif pan == 'right':
            right_volume = 1.
            left_volume = 0.
        elif pan < 0:
            right_volume = -1. / pan
            left_volume = 1.
        elif pan > 0:
            right_volume = 1.
            left_volume = 1. / pan
        else:
            return
        self._data[:, 0] *= left_volume
        self._data[:, 1] *= right_volume

    def _ms_to_samples(self, ms):

        return int(self._samplerate * (ms / 1000))

    def _set_duration(self, duration):

        if not duration:
            return
        self._data = self._data[:self._ms_to_samples(duration), :]

    def _set_fade_in(self, fade_in):

        if fade_in is None:
            return
        ramp = np.linspace(
            0, 1, self._ms_to_samples(fade_in)
        )[:self._data.shape[0]]
        self._data[:len(ramp), 0] *= ramp
        self._data[:len(ramp), 1] *= ramp

    @configurable
    def play(self, **playback_args):

        self._sound.stop()
        self._sound.play()
        if self.block:
            self.wait()

    def stop(self):

        self._sound.stop()

    def pause(self):

        self._sound.pause()

    def resume(self):

        self._sound.play()

    def is_playing(self):

        return self._sound.isPlaying

    def wait(self):

        while self._sound.isPlaying:
            self._keyboard.flush()

    @staticmethod
    def init_sound(experiment):

        global Sound, PLAYING

        from psychopy import prefs
        prefs.hardware['audioLib'] = [
            experiment.var.get('psycho_audiolib', DEFAULT_AUDIOLIB)]
        from psychopy import constants
        PLAYING = constants.PLAYING
        # Fixes a regression in psychopy introduced in
        # - https://github.com/psychopy/psychopy/commit/\
        #   45ed546b8e0a25ddb87156ef400687aaf31baf39
        # Should be removed as soon as this is fixed upstream
        import psychopy.sound._base
        psychopy.sound._base.defaultStim = []
        from psychopy.sound import Sound

    @staticmethod
    def close_sound(experiment):

        pass


# Non PEP-8 alias for backwards compatibility
psycho = Psycho



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from qtpy import QtWidgets
from .variable_inspector_widget import VariableInspectorWidget
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'variable_inspector', category=u'extension')


class VariableInspectorDockWidget(QtWidgets.QDockWidget):

    r"""A QDocktWidget that holds the variable inspector."""
    def __init__(self, main_window, ext):

        super().__init__(_(u'Variable inspector'), main_window)
        self.setWidget(VariableInspectorWidget(main_window, ext))
        self.setObjectName(u'variable_inspector')



"""Optionally repeat a cycle from a loop"""

category = "Flow control"
icon = "os-repeat_cycle"
controls = [
    {
        "label": "Repeat if",
        "tooltip": "A conditional expression that determines when the cycle is repeated",
        "type": "conditional_expression",
        "var": "condition",
        "verb": "repeat"
    }
]



"""Manages PyQode, the code-editor component"""

label = "PyQode manager"
priority = 101
settings = {
    "pyqode_font_size": 14,
    "pyqode_font_name": "Roboto Mono",
    "pyqode_color_scheme": "monokai",
    "pyqode_pyflakes_validation": True,
    "pyqode_pep8_validation": False,
    "pyqode_right_margin": False,
    "pyqode_highlight_caret_line": True,
    "pyqode_code_completion": True,
    "pyqode_code_completion_excluded_mimetypes": "markdown",
    "pyqode_code_folding": False,
    "pyqode_indentation": "auto",
    "pyqode_image_annotations": True,
    "pyqode_show_whitespaces": False,
    "pyqode_show_line_numbers": True,
    "pyqode_fixed_width": False,
    "pyqode_fixed_width_nchar": 80,
    "pyqode_autopep8": False,
    "pyqode_autopep8_lookbehind": 20,
    "pyqode_autopep8_aggressive": False,
    "pyqode_pep8_ignore": {
        "ide": "W191",
        "default": "W191"
    },
    "pyqode_pyflakes_ignore": {
        "ide": "",
        "default": "undefined name"
    },
    "pyqode_enable_breakpoints": {
        "ide": True,
        "default": False
    },
    "pyqode_tab_length": 4,
    "pyqode_line_wrap": True,
    "pyqode_shortcut_duplicate_line": "Ctrl+Shift+D",
    "pyqode_shortcut_swap_line_up": "Ctrl+Up",
    "pyqode_shortcut_swap_line_down": "Ctrl+Down",
    "pyqode_shortcut_toggle_breakpoint": "F12"
}
modes = ["default", "ide"]



"""This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from qtpy.QtWidgets import QStyledItemDelegate
from qtpy.QtGui import QColor, QFont, QPen
from qtpy.QtCore import Qt


class RunIfDelegate(QStyledItemDelegate):
    """Adds a comment next to run-if statements in the overview area to
    clarify that True means always run and False means never run.
    """
    def __init__(self, parent):
        super().__init__(parent)
        self._font = QFont('Roboto Condensed')
        self._pen = QPen(QColor('#78909c'))
        self._margin = 48
        self._width = None
        self._flags = Qt.AlignLeft | Qt.AlignVCenter
    
    def paint(self, painter, option, index):
        super().paint(painter, option, index)
        data = index.data()
        if data not in ('True', 'False'):
            return
        # The first time that we draw an annotation, we determine how far from
        # the left we need to draw it. This depends on the margin and the size
        # of the text
        if self._width is None:
            rect = painter.boundingRect(option.rect, 0, data)
            self._width = rect.right() + self._margin
        option.rect.setLeft(self._width)
        painter.setFont(self._font)
        painter.setPen(self._pen)
        painter.drawText(
            option.rect, self._flags,
            'Always run' if data == 'True' else 'Never run')



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
import os
import sys
import tarfile
from libopensesame import misc
from libopensesame.osexpfile import OSExpBase


class OSExpReader(OSExpBase):

    r"""A reader for osexp files of any format."""
    def __init__(self, exp, src):
        r"""Constructor. A side effect of calling this constructor is that all
        files in the file pool of src are extracted to the file pool folder of
        the experiment.

        Parameters
        ----------
        exp : experiment
            An experiment object.
        src : str
            An OpenSesame script, or the name of an OpenSesame experiment file.
        """
        super().__init__(exp)
        self._src = safe_decode(src)
        if self.format == u'script':
            self._read_script()
        elif self.format == u'scriptfile':
            self._read_scriptfile()
        elif self.format == u'targz':
            self._read_tarfile(format=u'r:gz')
        elif self.format == u'tar':
            self._read_tarfile(format=u'r')
        else:
            raise TypeError(u'"%s" is not an OpenSesame experiment')

    @property
    def script(self):
        """See osexpbase"""
        return self._script

    def _determine_format(self):
        """See osexpbase"""
        try:
            # Checking whether a path exists can appartently trigger a
            # ValueError when src is too long. Therefore, we use this awkward
            # construction.
            if not os.path.exists(self._src):
                raise ValueError(u'Path doesn\'t exist)')
        except ValueError:
            return u'script'
        try:
            tarfile.open(self._src, u'r:gz')
            return 'targz'
        except tarfile.ReadError:
            pass
        try:
            tarfile.open(self._src, u'r')
            return 'tar'
        except tarfile.ReadError:
            return 'scriptfile'

    def _read_script(self):
        r"""Reads a script (ie. not a file)"""
        self._script = self._src
        self._experiment_path = None

    def _read_scriptfile(self):
        r"""Reads a plain-text script file."""
        self._experiment_path = os.path.dirname(self._src)
        with safe_open(self._src) as fd:
            self._script = fd.read()

    def _read_tarfile(self, format):
        r"""Reads a tar or targz archive.

        Parameters
        ----------
        format : str
            The format to be used for tarfile.open().
        """
        self._experiment_path = os.path.dirname(self._src)
        tar = tarfile.open(self._src, format)
        for name in tar.getnames():
            # Tar doesn't return unicode in Python 2
            uname = safe_decode(name)
            folder, fname = os.path.split(uname)
            # Ignore everything except the pool folder
            if folder != u'pool':
                continue
            # Filenames are encoded with U+XXXX notation in the archive. This is
            # necessary to deal with absence of good unicode support in .tar.gz.
            fname = self._syntax.from_ascii(fname)
            pool_folder = safe_str(
                self._pool.folder(),
                enc=sys.getfilesystemencoding()
            )
            tar.extract(name, pool_folder)
            from_name = os.path.join(self._pool.folder(), uname)
            to_name = os.path.join(self._pool.folder(), fname)
            os.rename(from_name, to_name)
            os.rmdir(os.path.join(self._pool.folder(), folder))
        # Temporarily extract the script file to the file pool, and remove it
        # right away
        script_path = os.path.join(self._pool.folder(), u'script.opensesame')
        tar.extract(u'script.opensesame', self._pool.folder())
        with safe_open(script_path) as fd:
            self._script = fd.read()
        os.remove(script_path)



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from openexp import resources
from functools import partial
import inspect
try:
    from yamldoc._functiondoc import FunctionDoc
    from yamldoc import inherit as docinherit
except:
    FunctionDoc = None
    docinherit = type

_backend_info = None
_backend_types = [u'canvas', u'keyboard', u'mouse', u'sampler', u'color',
                  u'clock', u'log']


def backend_info():
    """
    arguments:
            experiment:
                    desc:	The experiment object.
                    type:	experiment

    returns:
            desc:	A dictionary with backend information.
            type:	dict
    """
    global _backend_info
    if _backend_info is None:
        with safe_open(resources['backend_info.yaml']) as fd:
            _backend_info = safe_yaml_load(fd.read())
    return _backend_info


def backend_guess(experiment, _type):
    r"""Guesses the backend of a specific type.

    Parameters
    ----------
    experiment : experiment
        The experiment object.
    _type : str
        A backend type, e.g. 'canvas'

    Returns
    -------
    str
        A backend, e.g. 'legacy'
    """
    if u'%s_backend' % _type in experiment.var:
        return experiment.var.get(u'%s_backend' % _type)
    d = backend_info()
    return d[experiment.var.canvas_backend][_type]


def backend_match(experiment):
    r"""Tries to determine which combination of backends is used by the
    experiment.

    Parameters
    ----------
    experiment : experiment
        The experiment object.

    Returns
    -------
    str
        The name of a backend combination, e.g. 'legacy'.
    """
    for name, info in backend_info().items():
        for _type in _backend_types:
            if backend_guess(experiment, _type) != info[_type]:
                break
        else:
            return name
    return u'custom'


def get_backend_mod(experiment, _type):
    r"""Gets a backend module.

    Parameters
    ----------
    experiment : experiment
        The experiment object.
    _type : str
        The backend type, e.g. u'canvas'.

    Returns
    -------
    module
        A module that contains the backend.
    """
    name = backend_guess(experiment, _type)
    return __import__('openexp._%s.%s' % (_type, name), fromlist=['dummy'])


def get_backend_class(experiment, _type):
    r"""Gets a backend class.

    Parameters
    ----------
    _type : str
        The backend type, e.g. u'canvas'.
    experiment : experiment
        The experiment object.

    Returns
    -------
    type
        A backend class.
    """
    name = backend_guess(experiment, _type)
    return getattr(get_backend_mod(experiment, _type), name)


def configurable(fnc):
    r"""A decorator that makes a function configurable.

    Parameters
    ----------
    fnc
        The function to make configurable.

    Returns
    -------
    A decorated configurable function.
    """
    def inner(self, *arglist, **kwdict):

        cfg = {}
        for key, val in list(kwdict.items()):
            if key in self.configurables:
                cfg[key] = val
                del kwdict[key]
        if len(cfg) > 0:
            old_cfg = self.get_config()
            self.set_config(**cfg)
        retval = fnc(self, *arglist, **kwdict)
        if len(cfg) > 0:
            self.set_config(**old_cfg)
        return retval
    # We need to copy the docstring and argument specification, otherwise using
    # this decorator will break the documentation functions.
    inner.__doc__ = fnc.__doc__
    inner.__argspec__ = inspect.getfullargspec(fnc)
    inner.__name__ = fnc.__name__
    if FunctionDoc is not None:
        inner._dict = FunctionDoc(fnc)._dict()
    return inner


def getter(key, self):
    r"""A getter function for configurable properties."""
    return self.__cfg__[key]


def setter(key, self, val):
    r"""A setter function for configurable properties."""
    self.set_config(**{key: val})


def deller():
    r"""A dummy delete function for configurable properties."""
    pass


def docstring(key):
    """
    returns:
            A docstring for configurable properties.
    """
    return """
	name:
		%(key)s
	desc:
		This is a read-and-write property. Changing the `%(key)s` property will
		affect subsequent operations.
	""" % {u'key': key}


class Backend:

    r"""A base backend that provides the configurable framework."""
    docinherit = type

    def __init__(self, configurables={}, **cfg):
        r"""Constructor.

        Parameters
        ----------
        configurables : dict, optional
            A dict of configurable properties with the name of the property as
            a key, and a validation function (which can be None) as value.
        **cfg : dict
            The initial values for all the configurable properties.
        """
        self.configurables = configurables
        for key in configurables:
            if hasattr(backend, key):
                continue
            fnc = property(
                partial(getter, key),
                partial(setter, key),
                deller,
                docstring(key))
            setattr(self.__class__, key, fnc)
        self.__cfg__ = {}
        _cfg = self.default_config().copy()
        _cfg.update(cfg)
        self.set_config(**_cfg)

    @classmethod
    def assert_list_or_None(cls, key, val):
        r"""Asserts that a value is a list or None.

        Parameters
        ----------
        key
            The name of the configurable.
        val
            The value of the configurable.
        """
        if not val is None and not isinstance(val, list):
            raise TypeError(
                u'%s should be a list or None, not %s' % (key, val))

    @classmethod
    def assert_bool(cls, key, val):
        r"""Asserts that a value is bool.

        Parameters
        ----------
        key
            The name of the configurable.
        val
            The value of the configurable.
        """
        if not isinstance(val, bool) and not isinstance(val, int):
            raise TypeError(
                u'%s should be True or False, not %s' % (key, val))

    @classmethod
    def assert_numeric_or_None(cls, key, val):
        r"""Asserts that a value is float, int, or None.

        Parameters
        ----------
        key
            The name of the configurable.
        val
            The value of the configurable.
        """
        if not val is None and not isinstance(val, int) and \
                not isinstance(val, float):
            raise TypeError(
                u'%s should be numeric (float or int) or None, not %s'
                % (key, val))

    @classmethod
    def assert_numeric(cls, key, val):
        r"""Asserts that a value is float or int.

        Parameters
        ----------
        key
            The name of the configurable.
        val
            The value of the configurable.
        """
        if not isinstance(val, int) and not isinstance(val, float):
            raise TypeError(
                u'%s should be numeric (float or int), not %s' % (key, val))

    @classmethod
    def assert_string(cls, key, val):
        r"""Asserts that a value is string or unicode.

        Parameters
        ----------
        key
            The name of the configurable.
        val
            The value of the configurable.
        """
        if not isinstance(val, str):
            raise TypeError(
                u'%s should be string (str or unicode), not %s' % (key, val))

    def get_config(self):
        """{nodoc}
        visible:	False

        returns:
                desc:	A dict that contains all configurables with names as keys
                                and values as values.
                type:	dict
        """
        return self.__cfg__.copy()

    def set_config(self, **cfg):
        r"""Updates the configurables.

        Parameters
        ----------
        **cfg : dict
            The to-be-updated configurables.
        """
        for key in cfg:
            if key not in self.configurables:
                raise RuntimeError(u'Unknown argument: %s' % key)
            assert_fnc = self.configurables[key]
            if assert_fnc is not None:
                assert_fnc(key, cfg[key])
        self.__cfg__.update(cfg)
        # Check whether the config is valid
        for key in self.configurables:
            if key not in self.__cfg__:
                raise RuntimeError(u'Invalid config: %s' % str(self.__cfg__))

    def default_config(self):
        """
        visible:	False

        returns:
                desc:	The default configurable values, with names as keys and
                                values as values.
                type:	dict
        """
        return {}


# Non PEP-8 alias for backwards compatibility
backend = Backend



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
import sys
from libopensesame.oslogging import oslogger
from libqtopensesame.misc.config import cfg
from pyqode.core import panels, modes
from pyqode.core.backend import server
from pyqode.core.api import (
    CodeEdit,
    Panel,
    CharBasedFoldDetector,
    IndentFoldDetector,
    ColorScheme
)
from ..modes import AutodetectIndentationMode

# For markdown it's annoying if quotes are autocompleted
DISABLE_AUTO_COMPLETE_QUOTES = ['markdown']
COMMENT_PREFIX = {
    'R': '# ',
    'javascript': '// ',
    'matlab': '% ',
    'octave': '% ',
    'perl': '# ',
    'php': '# ',
    'python': '# ',
    'ruby': '# ',
    'typescript': '// ',
    'yaml': '# ',
}


class FallbackCodeEdit(CodeEdit):

    # generic
    mimetypes = []

    #: the list of mimetypes that use char based fold detector
    _char_based_mimetypes = [
        'text/x-php',
        'text/x-c++hdr',
        'text/x-c++src',
        'text/x-chdr',
        'text/x-csrc',
        'text/x-csharp',
        'text/x-plain',
        'application/javascript'
    ]

    def __init__(self, parent=None):

        # PsychoPy deletes the _ built-in
        if '_' not in __builtins__:
            oslogger.warning('re-installing missing gettext built-in')
            import gettext
            gettext.NullTranslations().install()
        super(FallbackCodeEdit, self).__init__(parent)
        self._start_backend()
        self.show_whitespaces = cfg.pyqode_show_whitespaces
        if cfg.pyqode_image_annotations:
            self.panels.append(
                panels.ImageAnnotationsPanel(),
                panels.ImageAnnotationsPanel.Position.RIGHT
            )
        if cfg.pyqode_show_line_numbers:
            self.panels.append(panels.LineNumberPanel())
        self.panels.append(
            panels.SearchAndReplacePanel(),
            Panel.Position.BOTTOM
        )
        if cfg.pyqode_code_folding:
            self.panels.append(panels.FoldingPanel())
        self.modes.append(modes.CursorHistoryMode())
        self.modes.append(modes.ExtendedSelectionMode())
        self.modes.append(modes.CaseConverterMode())
        self.modes.append(modes.LineSorterMode())
        self.modes.append(modes.FileWatcherMode())
        if cfg.pyqode_highlight_caret_line:
            self.modes.append(modes.CaretLineHighlighterMode())
        self.modes.append(modes.PygmentsSyntaxHighlighter(
            self.document(),
            color_scheme=ColorScheme(cfg.pyqode_color_scheme))
        )
        self.modes.append(modes.ZoomMode())
        self.modes.append(modes.AutoIndentMode())
        self.modes.append(modes.IndenterMode())
        self.modes.append(modes.SymbolMatcherMode())
        self.modes.append(modes.SmartBackSpaceMode())
        self.modes.append(AutodetectIndentationMode())
        self.panels.append(panels.EncodingPanel(), Panel.Position.TOP)
        self.panels.append(panels.ReadOnlyPanel(), Panel.Position.TOP)
        if cfg.pyqode_fixed_width:
            self.panels.append(
                panels.MarginPanel(nchar=cfg.pyqode_fixed_width_nchar),
                Panel.Position.LEFT
            )
        if cfg.pyqode_right_margin:
            rmm = modes.RightMarginMode()
            self.modes.append(rmm)
            try:
                rmm.color = self.syntax_highlighter.formats['comment'] \
                    .foreground().color()
            except Exception as e:  # Don't know how safe this is
                pass
        self._set_comment_mode()

    def _start_backend(self):

        self._backend.start(
            server.__file__,
            sys.executable,
            reuse=True
        )

    def _set_comment_mode(self):

        comment_prefix = self.comment_prefix
        if comment_prefix:
            if 'CommentsMode' in self.modes.keys():
                self.modes.get('CommentsMode').prefix = comment_prefix
            else:
                self.modes.append(modes.CommentsMode(prefix=comment_prefix))
        elif 'CommentsMode' in self.modes.keys():
            self.modes.remove('CommentsMode')

    def setPlainText(self, txt, mime_type='', encoding=''):

        if not mime_type:
            mime_type = self.file.mimetype
        if not encoding:
            encoding = self.file.encoding
        if mime_type != u'text/x-plain':
            try:
                self.syntax_highlighter.set_lexer_from_mime_type(mime_type)
            except Exception:
                # The syntax_highlighter may not exist, or there may not be a lexer
                # for the mime type
                pass
            else:
                if mime_type in self._char_based_mimetypes:
                    self.syntax_highlighter.fold_detector = CharBasedFoldDetector()
                else:
                    self.syntax_highlighter.fold_detector = IndentFoldDetector()
        if self.language in DISABLE_AUTO_COMPLETE_QUOTES:
            self._disable_auto_complete_for_quotes()
        else:
            self._enable_auto_complete_for_quotes()
        super(FallbackCodeEdit, self).setPlainText(txt, mime_type, encoding)
        if (
                cfg.pyqode_code_completion and
                self.language not in cfg.pyqode_code_completion_excluded_mimetypes
        ):
            self.modes.append(modes.AutoCompleteMode())
            self.modes.append(modes.CodeCompletionMode())
            # The AutoIndentMode and SmartBackSpaceMode should catch key
            # presses after the # CodeCompletionMode, otherwise code
            # completions will have an ugly # newline in it. Toggling the
            # state accomplishes this, although # it's an ugly hack.
            for mode_name in ['AutoIndentMode', 'SmartBackSpaceMode']:
                if mode_name in self.modes.keys():
                    mode = self.modes.get(mode_name)
                    oslogger.debug('toggling mode {}'.format(mode_name))
                    if mode.enabled:
                        mode.enabled = False
                        mode.enabled = True
        self._set_comment_mode()

    def _enable_auto_complete_for_quotes(self):

        try:
            auto_complete_mode = self.modes.get('AutoCompleteMode')
        except KeyError:
            oslogger.debug('AutoCompleteMode not enabled')
            return
        auto_complete_mode.MAPPING['\''] = '\''
        auto_complete_mode.MAPPING['"'] = '"'

    def _disable_auto_complete_for_quotes(self):

        try:
            auto_complete_mode = self.modes.get('AutoCompleteMode')
        except KeyError:
            oslogger.debug('AutoCompleteMode not enabled')
            return
        if '\'' not in auto_complete_mode.MAPPING:
            return
        del auto_complete_mode.MAPPING['\'']
        del auto_complete_mode.MAPPING['"']

    def clone(self):

        clone = self.__class__(parent=self.parent())
        return clone

    @property
    def language(self):

        filenames = self.syntax_highlighter._lexer.filenames
        if u'*.R' in filenames:
            return u'R'
        if u'*.md' in filenames:
            return u'markdown'
        if u'*.js' in filenames:
            return u'javascript'
        return u'unknown'

    @property
    def comment_prefix(self):

        return COMMENT_PREFIX.get(self.language, None)

    def __repr__(self):

        return 'FallbackCodeEdit(path={}, language={})'.format(
            self.file.path,
            self.language
        )



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from pygame.locals import *
from openexp._mouse.mouse import Mouse
from openexp._coordinates.legacy import Legacy as LegacyCoordinates
from libopensesame.exceptions import UserAborted
from openexp.backend import configurable
import pygame


class Legacy(Mouse, LegacyCoordinates):

    r"""This is a mouse backend built on top of PyGame. For function
    specifications and docstrings, see `openexp._mouse.mouse`.
    """
    settings = {
        u"enable_escape": {
            u"name": u"Enable escape",
            u"description": u"Abort the experiment when the upper left and right corners are clicked",
            u"default": u"no",
        }
    }

    def __init__(self, experiment, **resp_args):

        Mouse.__init__(self, experiment, **resp_args)
        LegacyCoordinates.__init__(self)

    def set_pos(self, pos=(0, 0)):

        pygame.mouse.set_pos(self.to_xy(pos))

    def show_cursor(self, show=True):

        pygame.mouse.set_visible(show)
        Mouse.show_cursor(self, show=show)

    @configurable
    def get_click(self):

        return self._get_mouse_event(MOUSEBUTTONDOWN)

    @configurable
    def get_click_release(self):

        return self._get_mouse_event(MOUSEBUTTONUP)

    def _get_mouse_event(self, event_type):

        buttonlist = self.buttonlist
        timeout = self.timeout
        enable_escape = self.experiment.var.get(
            u'enable_escape',
            u'no',
            [u'yes', u'no']
        ) == u'yes'
        pygame.mouse.set_visible(self.visible)
        start_time = pygame.time.get_ticks()
        time = start_time
        while True:
            time = pygame.time.get_ticks()
            # Process the input
            for event in pygame.event.get():
                if event.type == KEYDOWN:
                    if event.key == pygame.K_ESCAPE:
                        self.experiment.pause()
                        continue
                    pygame.event.post(event)
                if event.type == event_type:
                    # Check escape sequence. If the top-left and top-right
                    # corner are clicked successively within 2000ms, the
                    # experiment is aborted
                    if (
                            enable_escape and event.pos[0] < 64
                            and event.pos[1] < 64
                    ):
                        _time = pygame.time.get_ticks()
                        while pygame.time.get_ticks() - _time < 2000:
                            for event in pygame.event.get():
                                if event.type == event_type:
                                    if (
                                            event.pos[0] > self.experiment.var.width-64
                                            and event.pos[1] < 64
                                    ):
                                        raise UserAborted(
                                            "The escape sequence was clicked/ tapped")
                    if buttonlist is None or event.button in buttonlist:
                        pygame.mouse.set_visible(self._cursor_shown)
                        return event.button, self.from_xy(event.pos), time
            if timeout is not None and time-start_time >= timeout:
                break
        pygame.mouse.set_visible(self._cursor_shown)
        return None, None, time

    def get_pos(self):

        pygame.event.get()
        return self.from_xy(pygame.mouse.get_pos()), self.experiment.clock.time()

    def get_pressed(self):

        pygame.event.get()
        return pygame.mouse.get_pressed()

    def flush(self):

        buttonclicked = False
        for event in pygame.event.get():
            if event.type == KEYDOWN and event.key == pygame.K_ESCAPE:
                self.experiment.pause()
            if event.type == MOUSEBUTTONDOWN:
                buttonclicked = True
        pygame.event.pump()
        return buttonclicked


# Non PEP-8 alias for backwards compatibility
legacy = Legacy


# -*- coding:utf-8 -*-

"""
This file is part of openexp.

openexp is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

openexp is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with openexp.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libqtopensesame.sketchpad_elements._base_line_arrow import BaseLineArrow
from libopensesame.sketchpad_elements import Arrow as ArrowRuntime


class Arrow(BaseLineArrow, ArrowRuntime):

    r"""An arrow element.
    See base_element for docstrings and function
    descriptions.
    """
    @classmethod
    def mouse_release(cls, sketchpad, pos):

        if cls.pos_start is None:
            return
        dx = pos[0] - cls.pos_start[0]
        dy = pos[1] - cls.pos_start[1]
        if abs(dx) < 1 and abs(dy) < 1:
            cls.pos_start = None
            sketchpad.canvas.removeItem(cls.preview)
            return
        properties = {
            u'x1':					cls.pos_start[0],
            u'y1':					cls.pos_start[1],
            u'x2':					pos[0],
            u'y2':					pos[1],
            u'color': 				sketchpad.current_color(),
            u'penwidth': 			sketchpad.current_penwidth(),
            u'arrow_head_width':	sketchpad.current_arrow_head_width(),
            u'arrow_body_width':	sketchpad.current_arrow_body_width(),
            u'arrow_body_length':	sketchpad.current_arrow_body_length(),
            u'fill':				sketchpad.current_fill(),
            u'show_if':				sketchpad.current_show_if()
        }
        e = Arrow(sketchpad, properties=properties)
        cls.pos_start = None
        sketchpad.canvas.removeItem(cls.preview)
        return e

    @staticmethod
    def requires_arrow_head_width():
        return True

    @staticmethod
    def requires_arrow_body_width():
        return True

    @staticmethod
    def requires_arrow_body_length():
        return True

    @staticmethod
    def requires_fill():
        return True


# Alias for backwards compatibility
arrow = Arrow



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from qtpy import QtCore, QtGui, QtWidgets
from libqtopensesame.misc.base_subcomponent import BaseSubcomponent
from libqtopensesame.misc import drag_and_drop
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'toolbar_items_item', category=u'core')


class ToolbarItemsItem(BaseSubcomponent, QtWidgets.QLabel):

    r"""A draggable toolbar icon."""
    def __init__(self, parent, item, pixmap=None):
        r"""Constructor.

        Parameters
        ----------
        parent : QWidget
            The parent.
        item : qtitem
            An item.
        pixmap : QPixmap
            A pixmap for the icon.
        """
        super().__init__(parent)
        self.setup(parent)
        self.item = item
        if pixmap is None:
            self.pixmap = self.theme.qpixmap(item)
        else:
            self.pixmap = pixmap
        # self.setMargin(6)
        self.setToolTip(_("Drag this <b>%s</b> item to the intended location in the overview area or into the item list of a sequence tab")
                        % self.item)
        self.setCursor(QtGui.QCursor(QtCore.Qt.OpenHandCursor))
        self.setPixmap(self.pixmap)

    def mousePressEvent(self, e):
        r"""Initiates a drag event after a mouse press.

        Parameters
        ----------
        e : QMouseEvent
            A mouse event.
        """
        if e.buttons() != QtCore.Qt.LeftButton:
            return
        name = u'new_%s' % self.item
        data = {
            u'type': u'item-snippet',
            u'main-item-name': name,
            u'items': [{
                u'item-type': self.item,
                u'item-name': name,
                u'script': u''
            }]
        }
        self.item_toolbar.collapse()
        drag_and_drop.send(self, data)


# Alias for backwards compatibility
toolbar_items_item = ToolbarItemsItem



# -*- coding:utf-8 -*-

"""
This file is part of openexp.

openexp is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

openexp is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with openexp.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libqtopensesame.misc.config import cfg
from libqtopensesame.sketchpad_elements._base_element import BaseElement


class BaseLineArrow(BaseElement):

    r"""A base class for line and arrow elements.
    See base_element for
    docstrings and function descriptions.
    """
    pos_start = None
    preview = None

    def set_pos(self, x=0, y=0):

        dx = self.properties[u'x2'] - self.properties[u'x1']
        dy = self.properties[u'y2'] - self.properties[u'y1']
        self.properties[u'x1'] = x
        self.properties[u'y1'] = y
        self.properties[u'x2'] = x+dx
        self.properties[u'y2'] = y+dy

    @classmethod
    def mouse_press(cls, sketchpad, pos):

        if cls.pos_start is not None:
            return
        cls.pos_start = pos
        cls.preview = sketchpad.canvas.line(pos[0], pos[1], pos[0], pos[1],
                                            color=cfg.sketchpad_preview_color,
                                            penwidth=cfg.sketchpad_preview_penwidth)

    @classmethod
    def mouse_move(cls, sketchpad, pos):

        if cls.pos_start is None or cls.preview is None:
            return
        cls.preview.setLine(cls.pos_start[0], cls.pos_start[1], pos[0], pos[1])

    @classmethod
    def reset(cls):

        cls.pos_start = None

    @staticmethod
    def requires_color():
        return True

    @staticmethod
    def requires_penwidth():
        return True


# Alias for backwards compatibility
base_line_arrow = BaseLineArrow



"""Sends a bug report when an exception occurs"""
 
icon = "applications-accessories"
label = "Bug report"
priority = 999999
settings = {
  "bug_report_url": "http://www.cogsci.nl/bug.php"
}
modes = ["default", "ide"]



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from openexp._canvas._element.element import Element


class Polygon(Element):

    def __init__(self, canvas, vertices, **properties):

        properties = properties.copy()
        properties.update({'vertices': vertices})
        self._shapely_polygon = None
        Element.__init__(self, canvas, **properties)

    @property
    def rect(self):

        left = min(x for x, y in self.vertices)
        right = max(x for x, y in self.vertices)
        top = min(y for x, y in self.vertices)
        bottom = max(y for x, y in self.vertices)
        return left, top, right-left, bottom-top

    def __contains__(self, xy):

        # Shapely is used to determine whether a point falls exactly within the
        # polygon. If shapely isn't available, we fall back to a simple bounding
        # box. The shapely polygon is stored for performance.
        try:
            from shapely.geometry import Polygon as ShapelyPolygon, Point
        except ImportError:
            return Element.__contains__(self, xy)
        if self._shapely_polygon is None:
            self._shapely_polygon = ShapelyPolygon(self.vertices)
        return self._shapely_polygon.contains(Point(*xy))

    def _on_attribute_change(self, **kwargs):

        # When a change occurs the shapely polygon (if any) needs to be
        # redetermined in __contains__
        self._shapely_polygon = None
        Element._on_attribute_change(self, **kwargs)



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.sketchpad_elements._base_element import BaseElement


class Image(BaseElement):

    r"""An image element for the sketchpad."""
    def __init__(self, sketchpad, string):
        r"""Constructor.

        Parameters
        ----------
        sketchpad
            A sketchpad object.
        string
            A definition string.
        """
        defaults = [
            (u'x', None),
            (u'y', None),
            (u'file', None),
            (u'scale', 1),
            (u'rotation', 0),
            (u'center', 1),
        ]
        super().__init__(sketchpad, string, defaults=defaults)

    def draw(self):
        r"""Draws the element to the canvas of the sketchpad."""
        properties = self.eval_properties()
        try:
            _file = self.pool[properties[u'file']]
        except:
            _file = u''
        return self.canvas.image(_file,
                                 center=properties[u'center'] == 1,
                                 x=properties[u'x'],
                                 y=properties[u'y'],
                                 scale=properties[u'scale'],
                                 rotation=properties[u'rotation']
                                 )

    def to_string(self):
        """See base_element."""
        # For backwards compatibility we don't include the rotation keyword if
        # it is left on 0 (i.e. no rotation). This is necessary for sketchpads
        # to be parse-able in 3.1.X.
        return self.syntax.create_cmd(u'draw', [self._type],
            {
                var: val for var, val in self.properties.items()
                if var not in (u'name', u'rotation') or val
            }
        )


# Alias for backwards compatibility
image = Image



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
import re
import webcolors
import colorsys
import numbers
from libopensesame.exceptions import MissingDependency, InvalidColor

RGB_HEX6 = r'#(?P<r>[0-9a-fA-F]{2})(?P<g>[0-9a-fA-F]{2})(?P<b>[0-9a-fA-F]{2})$'
RGB_HEX3 = r'#(?P<r>[0-9a-fA-F])(?P<g>[0-9a-fA-F])(?P<b>[0-9a-fA-F])$'
RGB_255 = r'rgb\(\s*(?P<r>\d+)\s*,\s*(?P<g>\d+)\s*,\s*(?P<b>\d+)\s*\)\s*$'
RGB_PERC = r'rgb\(\s*(?P<r>[+-]?(\d+(\.\d*)?|\.\d+)([eE][+-]?\d+)?)%\s*,\s*(?P<g>[+-]?(\d+(\.\d*)?|\.\d+)([eE][+-]?\d+)?)%\s*,\s*(?P<b>[+-]?(\d+(\.\d*)?|\.\d+)([eE][+-]?\d+)?)%\s*\)\s*$'
HSL = r'hsl\(\s*(?P<h>[+-]?(\d+(\.\d*)?|\.\d+)([eE][+-]?\d+)?)\s*,\s*(?P<s>[+-]?(\d+(\.\d*)?|\.\d+)([eE][+-]?\d+)?)%\s*,\s*(?P<l>[+-]?(\d+(\.\d*)?|\.\d+)([eE][+-]?\d+)?)%\s*\)\s*$'
HSV = r'hsv\(\s*(?P<h>[+-]?(\d+(\.\d*)?|\.\d+)([eE][+-]?\d+)?)\s*,\s*(?P<s>[+-]?(\d+(\.\d*)?|\.\d+)([eE][+-]?\d+)?)%\s*,\s*(?P<v>[+-]?(\d+(\.\d*)?|\.\d+)([eE][+-]?\d+)?)%\s*\)\s*$'
LAB = r'lab\(\s*(?P<l>[+-]?(\d+(\.\d*)?|\.\d+)([eE][+-]?\d+)?)\s*,\s*(?P<a>[+-]?(\d+(\.\d*)?|\.\d+)([eE][+-]?\d+)?)\s*,\s*(?P<b>[+-]?(\d+(\.\d*)?|\.\d+)([eE][+-]?\d+)?)\s*\)\s*$'


def _is_rgb(colorspec):

    return (
        hasattr(colorspec, '__len__')
        and len(colorspec) == 3
        and all(
            isinstance(i, numbers.Integral)
            and 0 <= i <= 255
            for i in colorspec
        )
    )


class Color:

    r"""Converts various color specifications to a back-end specific format.
    Valid color specifications are described in more detail in
    openexp._canvas.canvas.canvas.
    """
    def __init__(self, experiment, colorspec):
        r"""Constructor.

        Parameters
        ----------
        experiment : experiment
            The experiment object.
        colorspec : str, unicode, tuple, int
            A color specification.
        """
        self.experiment = experiment
        self.colorspec = colorspec
        self.hexcolor = self.to_hex(self.colorspec)
        self.backend_color = self.to_backend_color(self.hexcolor)

    def __repr__(self):
        """
        returns:
                A representation of the color, which matches the color
                specification.
        """
        return self.colorspec

    @staticmethod
    def to_hex(colorspec):
        r"""Converts a color specification to a seven-character lowercase
        hexadecimal color string, such as '#ff0000'.

        Parameters
        ----------
        colorspec : str, unicode, array-like, int
            A color specification.

        Returns
        -------
        unicode
            A hexadecimal color specification.
        """
        if isinstance(colorspec, float):
            colorspec = int(colorspec)
        if isinstance(colorspec, int):  # 0-255 luminance value
            return webcolors.rgb_to_hex((colorspec, colorspec, colorspec))
        if _is_rgb(colorspec):
            return webcolors.rgb_to_hex(colorspec)
        if hasattr(colorspec, 'hexcolor'):
            return colorspec.hexcolor
        if not isinstance(colorspec, str):
            raise InvalidColor(colorspec)
        try:  # 0-255 luminance value passed as string
            colorspec = int(colorspec)
        except ValueError:
            pass
        else:
            return webcolors.rgb_to_hex((colorspec, colorspec, colorspec))
        try:
            return webcolors.name_to_hex(colorspec)
        except ValueError:
            pass
        m = re.match(RGB_HEX6, colorspec)
        if m:
            return colorspec.lower()
        m = re.match(RGB_HEX3, colorspec)
        if m:
            return webcolors.rgb_to_hex((
                int(m.group('r') * 2, base=16),
                int(m.group('g') * 2, base=16),
                int(m.group('b') * 2, base=16)
            ))
        m = re.match(RGB_255, colorspec)
        if m:
            return webcolors.rgb_to_hex(
                (int(m.group('r')), int(m.group('g')), int(m.group('b')))
            )
        m = re.match(RGB_PERC, colorspec)
        if m:
            return webcolors.rgb_percent_to_hex(
                (m.group('r'), m.group('g'), m.group('b'))
            )
        m = re.match(HSL, colorspec)
        if m:
            # RGB values between 0 and 1
            r, g, b = colorsys.hls_to_rgb(
                float(m.group('h')) / 360,
                float(m.group('l')) / 100,
                float(m.group('s')) / 100
            )
            return webcolors.rgb_to_hex(
                (int(r * 255), int(g * 255), int(b * 255))
            )
        m = re.match(HSV, colorspec)
        if m:
            # RGB values between 0 and 1
            r, g, b = colorsys.hsv_to_rgb(
                float(m.group('h')) / 360,
                float(m.group('s')) / 100,
                float(m.group('v')) / 100
            )
            return webcolors.rgb_to_hex(
                (int(r * 255), int(g * 255), int(b * 255))
            )
        m = re.match(LAB, colorspec)
        if m:
            try:
                from psychopy.tools import colorspacetools as cst
            except ImportError:
                raise MissingDependency(
                    'CIE L*a*b* color space requires PsychoPy')
            # RGB values are between -1 and 1
            r, g, b = cst.cielab2rgb(
                (
                    float(m.group('l')),
                    float(m.group('a')),
                    float(m.group('b'))
                ),
                transferFunc=cst.srgbTF
            )
            return webcolors.rgb_to_hex((
                int((r + 1) * 127.5),
                int((g + 1) * 127.5),
                int((b + 1) * 127.5),
            ))
        raise InvalidColor(colorspec)

    def to_backend_color(self, hexcolor):
        r"""Converts a hexadecimal color string to a backend-specific color
        object.

        Parameters
        ----------
        hexcolor : str, unicode
            A hexadecimal color specification.

        Returns
        -------
        A backend-specific color object.
        """
        return hexcolor


# Non PEP-8 alias for backwards compatibility
color = Color



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.exceptions import InvalidKeyName


CHAR_TO_NAME = {
    u'`': u'grave',
    u'~': u'asciitilde',
    u'^': u'asciicircum',
    u'%': u'percent',
    u'!': u'exclamation',
    u'"': u'doublequote',
    u'#': u'hash',
    u'$': u'dollar',
    u'&': u'ampersand',
    u'\'': u'apostrophe',
    u'*': u'asterisk',
    u'+': u'plus',
    u',': u'comma',
    u'-': u'minus',
    u'.': u'period',
    u'/': u'slash',
    u':': u'colon',
    u';': u'semicolon',
    u'=': u'equal',
    u'?': u'question',
    u'@': u'at',
    u'\\': u'backslash',
    u'|': u'bar',
    u'<': u'less',
    u'>': u'greater',
    u'[': u'bracketleft',
    u']': u'bracketright',
    u'(': u'parenleft',
    u')': u'parenright',
    u'{': u'braceleft',
    u'}': u'braceright',
    u'_': u'underscore'
}
NAME_TO_CHAR = {name: char for char, name in CHAR_TO_NAME.items()}


class KeyBabel:

    r"""Converts between different ways to represent keys through numeric
    values (keycodes), different names (strings), and None (for timeout).
    """
    def __init__(self, keyboard):
        r"""Constructor.

        Parameters
        ----------
        keyboard
            A Keyboard object
        """
        self._keyboard = keyboard

    def standard_name(self, key, shift=False):
        r"""Gives the standard name for a key. This is the shortest name, so it
        corresponds to a character if a character is available.

        Parameters
        ----------
        key
            The key to standardize.
        shift, optional
            Indicates whether the name should be cast to uppercase (True) or
            lowercase (False)

        Returns
        -------
        The standard name for the key.
        """
        short_name = (
            key
            if isinstance(key, str) and len(key) == 1
            else sorted(self.synonyms(key), key=len)[0]
        )
        return short_name.upper() if shift else short_name.lower()

    def synonyms(self, key):
        r"""Gives a set of synonyms for a key.

        Parameters
        ----------
        key
            The key to find synonyms for.

        Returns
        -------
        A set of synonyms.
        """
        if key is None:
            return self._none_synonyms()
        if isinstance(key, int):
            return self._keycode_synonyms(key)
        if isinstance(key, str):
            return self._str_synonyms(key)
        raise InvalidKeyName(key)

    def _none_synonyms(self):

        return {None, 'None', 'none', 'NONE'}

    def _keycode_synonyms(self, key):

        keyname = self._keyboard._keycode_to_str(key)
        return {key, keyname.upper(), keyname.lower()}

    def _str_synonyms(self, key):

        key_lower = key.lower()
        synonyms = {key, key.upper(), key_lower}
        if key_lower in CHAR_TO_NAME:
            # Add 'exclamation' for '!'
            synonyms.add(CHAR_TO_NAME[key_lower])
        elif key_lower in NAME_TO_CHAR:
            # Add '!' for 'exclamation'
            synonyms.add(NAME_TO_CHAR[key_lower])
        # Add timeout as NoneType
        if key_lower == u'none':
            synonyms.add(None)
        return synonyms



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
import fnmatch
import subprocess
from libopensesame.oslogging import oslogger
from libqtopensesame.misc.config import cfg
from libqtopensesame.extensions import BaseExtension


class FileHandler(BaseExtension):

    r"""Opens files in external applications."""
    def event_startup(self):

        # First make sure that the bindings config evaluates to a dict
        try:
            self._bindings = safe_yaml_load(cfg.external_bindings)
            if not isinstance(self._bindings, dict):
                raise TypeError('YAML is not a dict')
        except Exception as e:
            oslogger.error('invalid YAML in {}'.format(cfg.external_bindings))
            self._bindings = {}
        # And then take only those entries where both the pattern and the
        # command are strings
        self._bindings = {
            pattern: cmd
            for pattern, cmd in self._bindings.items()
            if isinstance(pattern, str) and isinstance(cmd, str)
        }

    def _is_experiment(self, path):

        path = path.lower()
        return (
            path.endswith(u'.osexp') or
            path.endswith(u'.opensesame') or
            path.endswith(u'.opensesame.tar.gz')
        )

    def provide_file_handler(self, path):

        if not isinstance(path, str) or self._is_experiment(path):
            return
        for pattern, cmd in self._bindings.items():
            if fnmatch.fnmatch(path, pattern):
                oslogger.debug('providing file handler for {}'.format(path))
                return self._file_handler
        oslogger.debug('not providing file handler for {}'.format(path))

    def _file_handler(self, path):

        for pattern, cmd in self._bindings.items():
            if fnmatch.fnmatch(path, pattern):
                try:
                    subprocess.Popen([cmd, path])
                except Exception as e:
                    self.notify(_(
                        u'Failed to open file. '
                        u'See debug window for error message.'
                    ))
                    self.console.write(e)



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from openexp import backend
from libopensesame.oslogging import oslogger
import tempfile
import os

# A list of temporary files that should be cleaned up.
temp_files = []


def Canvas(experiment, *arglist, **kwdict):
    r"""A factory that returns a back-end specific canvas object.

    Parameters
    ----------
    experiment : experiment
        The experiment object.
    *arglist : list
        See canvas.__init__().
    **kwdict : dict
        See canvas.__init__().
    """
    cls = backend.get_backend_class(experiment, u'canvas')
    return cls(experiment, *arglist, **kwdict)


def init_display(experiment):
    r"""Calls the back-end specific init_display function.

    Parameters
    ----------
    experiment
        The experiment object.
    type
        experiment
    """
    cls = backend.get_backend_class(experiment, u'canvas')
    cls.init_display(experiment)


def close_display(experiment):
    r"""Calls the back-end specific close_display function.

    Parameters
    ----------
    experiment
        The experiment object.
    type
        experiment
    """
    cls = backend.get_backend_class(experiment, u'canvas')
    cls.close_display(experiment)


def clean_up(verbose=False):
    r"""Cleans up temporary pool folders.

    Parameters
    ----------
    verbose : bool, optional
        Indicates if debugging output should be provided.
    """
    global temp_files
    if verbose:
        oslogger.debug(u"cleaning up temporary pool folders")
    for path in temp_files:
        if verbose:
            oslogger.debug(u"removing '%s'" % path)
        try:
            os.remove(path)
        except Exception as e:
            if verbose:
                oslogger.error(u"failed to remove '%s': %s" % (path, e))


def gabor_file(*arglist, **kwdict):
    r"""Creates a temporary file containing a Gabor patch.

    Parameters
    ----------
    *arglist : list
        See canvas.gabor() for a description of arguments.
    **kwdict : dict
        See canvas.gabor() for a description of keywords.


    Returns
    -------
    A path to the image file.
    """
    from openexp._canvas import canvas as _canvas
    fd, fname = tempfile.mkstemp(suffix=u'.png')
    temp_files.append(fname)
    im = _canvas._gabor(*arglist, **kwdict)
    im.save(fname)
    os.close(fd)
    return fname


def noise_file(*arglist, **kwdict):
    r"""Creates a temporary file containing a noise patch.

    Parameters
    ----------
    *arglist : list
        See canvas.noise_path() for a description of arguments.
    **kwdict : dict
        See canvas.noise_path() for a description of keywords.


    Returns
    -------
    A path to the image file.
    """
    from openexp._canvas import canvas as _canvas
    fd, fname = tempfile.mkstemp(suffix=u'.png')
    temp_files.append(fname)
    im = _canvas._noise_patch(*arglist, **kwdict)
    im.save(fname)
    os.close(fd)
    return fname


# Non PEP-8 alias for backwards compatibility
canvas = Canvas



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libqtopensesame.misc import config
from libqtopensesame.widgets.base_widget import BaseWidget
from qtpy import QtWidgets, QtCore
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'unused_widget', category=u'core')


class UnusedWidget(BaseWidget):

    r"""The unused items widget."""
    tab_name = u'__unused__'

    def __init__(self, main_window):
        r"""Constructor.

        Parameters
        ----------
        main_window : qtopensesame
            The main-window object.
        """
        super().__init__(main_window)
        header_hbox = QtWidgets.QHBoxLayout()
        header_hbox.addWidget(self.theme.qlabel(u"unused"))
        header_label = QtWidgets.QLabel()
        header_label.setText(_(u"<b><font size='5'>Unused</font></b>"))
        header_hbox.addWidget(header_label)
        header_hbox.addStretch()
        self.purge_button = QtWidgets.QPushButton(self.theme.qicon(u"purge"),
                                                  _(u"Permanently delete unused items"))
        self.purge_button.setIconSize(QtCore.QSize(16, 16))
        self.purge_button.clicked.connect(self.purge_unused)
        purge_hbox = QtWidgets.QHBoxLayout()
        purge_hbox.addWidget(self.purge_button)
        purge_hbox.addStretch()
        purge_widget = QtWidgets.QWidget()
        purge_widget.setLayout(purge_hbox)
        vbox = QtWidgets.QVBoxLayout()
        vbox.addWidget(purge_widget)
        vbox.addStretch()
        self.setLayout(vbox)
        self.__unused_tab__ = True

    def purge_unused(self):
        r"""Purges all unused items."""
        resp = QtWidgets.QMessageBox.question(self.main_window.ui.centralwidget,
            _(u"Permanently delete items?"),
            _(u"Are you sure you want to permanently delete all unused items? This action cannot be undone."),
            QtWidgets.QMessageBox.Yes, QtWidgets.QMessageBox.No)
        if resp == QtWidgets.QMessageBox.No:
            return
        self.extension_manager.fire(u'prepare_purge_unused_items')
        self.overview_area.locked = True
        for item_name in self.experiment.items.unused():
            self.experiment.items[item_name].close_tab()
            del self.experiment.items[item_name]
        self.overview_area.locked = False
        self.experiment.build_item_tree()
        self.purge_button.setDisabled(True)
        self.extension_manager.fire(u'purge_unused_items')

    def on_activate(self):
        r"""Is called when the widget becomes visible."""
        self.purge_button.setDisabled(len(self.experiment.items.unused()) == 0)


# Alias for backwards compatibility
unused_widget = UnusedWidget



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libqtopensesame.misc.base_subcomponent import BaseSubcomponent
from qtpy import QtCore, QtWidgets


class BaseDialog(QtWidgets.QDialog, BaseSubcomponent):

    r"""A base class for dialogs."""
    def __init__(self, main_window, ui=None, *arglist, **kwdict):
        """
        desc:
                Constructor.

        arguments:
                main_window:	A qtopensesame object.

        keywords:
                ui:
                                                An id for a user-interface file, for example
                                                'dialogs.quick_switcher'.

        argument-list:
        *arglist:			A list of arguments to be passed onto QDialog.

        keyword-dict:
        *kwdict:			A dict of keywords to be passed onto QDialog.
        """
        super().__init__(main_window, *arglist, **kwdict)
        self.setup(main_window, ui=ui)


# Alias for backwards compatibility
base_dialog = BaseDialog



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.item import Item
from . import ItemTask, InlineTask


class Coroutines(Item):

    def reset(self):
        """See item."""
        self.var.duration = 5000
        self.var.flush_keyboard = 'yes'
        self.var.function_name = ''
        self.var.end_after_item = ''
        self.schedule = []
        self._events = []
        self.pre_cycle_functions = []
        self.post_cycle_functions = []

    def event(self, msg):

        self._events.append((self.clock.time(), msg))

    @property
    def event_log(self):

        return '\n'.join(['%d: %s' % (t, msg) for t, msg in self._events])

    def is_coroutine(self, item_name):
        return hasattr(self.experiment.items[item_name], 'coroutine')

    def is_oneshot_coroutine(self, item_name):
        try:
            return self.experiment.items[item_name].is_oneshot_coroutine
        except AttributeError:
            return False

    def from_string(self, string):
        self.var.clear()
        self.comments = []
        self.reset()
        if string is None:
            return
        for s in string.split('\n'):
            self.parse_variable(s)
            # run item_name start=1000 end=2000 runif="always"
            cmd, arglist, kwdict = self.syntax.parse_cmd(s)
            if cmd != 'run' or not len(arglist):
                continue
            item_name = arglist[0]
            start_time = kwdict.get('start', 0)
            end_time = kwdict.get('end', 0)
            cond = kwdict.get('runif', 'always')
            enabled = kwdict.get('enabled', 'yes') == 'yes'
            self.schedule.append((item_name, start_time, end_time, cond, enabled))

    def to_string(self):
        s = Item.to_string(self)
        for item_name, start_time, end_time, cond, enabled in self.schedule:
            # If the item doesn't exist yet, then we simply go with the times
            # from the schedule. This happens during loading, if the
            # coroutines script is parsed before the scripts of the items that
            # are in it.
            if (item_name in self.experiment.items and 
                    self.is_oneshot_coroutine(item_name)):
                end_time = start_time
            s += '\t' + self.syntax.create_cmd('run', [item_name], {
                'start': start_time,
                'end': end_time,
                'runif': cond,
                'enabled': 'yes' if enabled else 'no'
            }) + '\n'
        return s

    def prepare(self):
        super().prepare()
        self.event('prepare coroutines')
        self._schedule = []
        for item_name, start_time, end_time, cond, enabled in self.schedule:
            if not enabled:
                continue
            if not self.python_workspace._eval(self.syntax.compile_cond(cond)):
                continue
            t = ItemTask(
                self, self.experiment.items[item_name],
                self.syntax.auto_type(self.syntax.eval_text(start_time)),
                self.syntax.auto_type(self.syntax.eval_text(end_time)),
                abort_on_end=item_name == self.var.end_after_item)
            self._schedule.append(t)
        if self.var.function_name != u"":
            t = InlineTask(self, self.var.function_name, self.python_workspace,
                           0, self.var.duration)
            self._schedule.append(t)

    def run(self):
        # Launch all coroutines
        for task in self._schedule:
            task.launch()
        self._schedule.sort(key=lambda task: task.start_time)
        dt = 0
        active = []
        t0 = self.clock.time()
        i = 0
        running = True
        while running and dt < self.var.duration:
            # Activate coroutines by start time
            while self._schedule and self._schedule[0].started(dt):
                active.append(self._schedule.pop(0))
                active.sort(key=lambda task: task.end_time)
            for fnc in self.pre_cycle_functions:
                fnc()
            # Run all active coroutines. If a task returns alive, it should be
            # kept as an active task; if it returns DEAD, it should be removed
            # from the active tasks; if it returns ABORT, the whole coroutines
            # should be aborted.
            _active = []
            for task in active:
                status = task.step()
                if status == task.ALIVE:
                    _active.append(task)
                    continue
                if status == task.ABORT:
                    running = False
            active = _active
            for fnc in self.post_cycle_functions:
                fnc()
            # De-activate coroutines by end time
            while active and active[0].stopped(dt):
                active.pop(0)
            dt = self.clock.time()-t0
            i += 1
        self.event('killed after %d ms' % (self.clock.time()-t0))
        # Kill pending coroutines
        for task in active:
            task.kill()
        self.event('trampoline took %d ms' % (self.clock.time()-t0))
        self.experiment.var.coroutines_cycles = i
        self.experiment.var.coroutines_duration = dt
        self.experiment.var.coroutines_mean_cycle_duration = 1.*dt/i
        self.event('%d cycles with an average duration of %.4f ms' %
                   (
                       self.experiment.var.coroutines_cycles,
                       self.experiment.var.coroutines_mean_cycle_duration
                   )
                   )

    def var_info(self):
        l = []
        l.append((u"coroutines_cycles", u"[Determined at runtime]"))
        l.append((u"coroutines_duration", u"[Determined at runtime]"))
        l.append((u"coroutines_mean_cycle_duration",
                 u"[Determined at runtime]"))
        return l



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
import pygame
from openexp._canvas import legacy
from openexp._canvas._text.text import Text
from openexp._canvas._element.legacy import LegacyElement
from openexp import resources


class Legacy(LegacyElement, Text):

    def __init__(self, canvas, text, x, y, **properties):

        self._antialias = True
        Text.__init__(self, canvas, text, x, y, **properties)
        self._set_font()

    def prepare(self):

        self._set_font()
        surface = self._font.render(self.text, self._antialias,
                                    self.color.backend_color)
        self.surface.blit(surface, self.to_xy(self.x, self.y))

    @property
    def size(self):

        return self._font.size(self.text)

    def _set_font(self):

        self._font = self._pygame_font(self.experiment, self.font_family,
                                       self.font_size)
        self._font.set_bold(self.font_bold)
        self._font.set_italic(self.font_italic)
        self._font.set_underline(self.font_underline)

    @staticmethod
    def _pygame_font(experiment, family, size):

        if (family, size) in legacy.fonts:
            return legacy.fonts[(family, size)]
        try:
            path = resource[f'{family}.ttf']
        except:
            # If the family cannot be found in the filepool, assume that it is
            # a system font.
            font = pygame.font.SysFont(family, size)
        else:
            fd = open(path, u'rb')
            legacy.fileobjects.append(fd)
            font = pygame.font.Font(fd, size)
        legacy.fonts[(family, size)] = font
        return font



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
import os
import pygame
from libopensesame.exceptions import UnsupportedImageFormat, ImageDoesNotExist
from openexp._canvas._image.image import Image
from openexp._canvas._element.legacy import LegacyElement


class Legacy(LegacyElement, Image):

    def prepare(self):

        if not hasattr(self, '_image_surface') or self._dirty:
            self._dirty = False
            fname = safe_decode(self.fname)
            if not os.path.isfile(fname):
                raise ImageDoesNotExist(fname)
            with open(fname, u'rb') as fd:
                try:
                    self._image_surface = pygame.image.load(fd)
                except pygame.error:
                    raise UnsupportedImageFormat(fname)
            # After rotation, the figure gets bigger. We therefore need to
            # compensate by moving it a bit
            if self.rotation is not None and self.rotation != 0:
                w1, h1 = self._image_surface.get_size()
                self._image_surface = pygame.transform.rotate(
                    self._image_surface.convert_alpha(), -self.rotation
                )
                w2, h2 = self._image_surface.get_size()
                self._dx = (w2-w1)/2
                self._dy = (h2-h1)/2
            else:
                self._dx = self._dy = 0
            if self.scale is not None:
                try:
                    self._image_surface = pygame.transform.smoothscale(
                        self._image_surfacesurface,
                        (int(self._image_surface.get_width()*self.scale),
                         int(self._image_surface.get_height()*self.scale))
                    )
                except:
                    self._image_surface = pygame.transform.scale(
                        self._image_surface,
                        (int(self._image_surface.get_width()*self.scale),
                         int(self._image_surface.get_height()*self.scale))
                    )
                self._dx *= self.scale
                self._dy *= self.scale
        size = self._image_surface.get_size()
        x, y = self.to_xy(self.x, self.y)
        if self.center:
            x -= size[0] / 2
            y -= size[1] / 2
        else:
            x -= self._dx
            y -= self._dy
        self.surface.blit(self._image_surface, (x, y))

    @staticmethod
    def _setter(key, self, val):

        self._dirty = True
        Image._setter(key, self, val)



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.exceptions import InvalidValue


class ResponseInfo:

    r"""A single response."""
    def __init__(self, response_store, response=None, correct=None,
                 response_time=None, item=None, feedback=True):

        self._response_store = response_store

        # Sanity checks
        if not isinstance(response_time, (int, float)) and \
                response_time is not None:
            raise InvalidValue(f'response_time should be a numeric value or '
                               f'None, not {response_time}')
        if response_time is None:
            self.response_time = None
        else:
            self.response_time = float(response_time)
        self.response = response
        self.item = item
        self.feedback = feedback
        if correct not in (0, 1, True, False, None):
            raise InvalidValue(
                f'correct should be 0, 1, True, False, or None, not {correct}')
        if correct is None:
            self.correct = None
        else:
            self.correct = int(correct)

    def match(self, **kwdict):

        for key, val in kwdict.items():
            if getattr(self, key) != val:
                return False
        return True

    def matchnot(self, **kwdict):

        for key, val in kwdict.items():
            if getattr(self, key) == val:
                return False
        return True

    def __str__(self):

        return u'response(response=%s,response_time=%s,correct=%s,item=%s,feedback=%s)' \
            % (self.response, self.response_time, self.correct, self.item,
               self.feedback)


class ResponseStore:
    """The `responses` object contains the history of the responses that were
    collected during the experiment.

    A `responses` object is created automatically when the experiment starts.

    In addition to the functions listed below, the following semantics are
    supported:

    __Example__

    ~~~ .python
    # Loop through all responses, where last-given responses come first
    # Each response has correct, response, response_time, item, and feedback
    # attributes.
    for response in responses:
        print(response.correct)
    # Print the two last-given responses
    print('last_two responses:')
    print(responses[:2])
    ~~~

    [TOC]
    """
    def __init__(self, experiment):
        r"""Constructor.

        Parameters
        ----------
        experiment : experiment.
            The experiment object.
        """
        self._experiment = experiment
        self._responses = []
        self._feedback_from = 0

    @property
    def acc(self):
        r"""The percentage of correct responses for all responses that are
        included in feedback. If there are no responses to give feedback on,
        'undefined' is returned.

        Examples
        --------
        >>> print('The accuracy was %s%%' % responses.acc)
        """
        l = self._select(feedback=True)._selectnot(correct=None).correct
        if not l:
            return u'undefined'
        return int(100 * sum(l) / len(l))

    @property
    def avg_rt(self):
        r"""The average response time for all responses that are included in
        feedback. If there are no responses to give feedback on, 'undefined' is
        returned.

        Examples
        --------
        >>> print('The average RT was %s ms' % responses.avg_rt)
        """
        l = self._select(feedback=True)._selectnot(response_time=None) \
                .response_time
        if not l:
            return u'undefined'
        return int(sum(l) / len(l))

    @property
    def response(self):
        r"""A list of all response values. (I.e. not response objects, but
        actual response keys, buttons, etc.)

        Examples
        --------
        >>> for response in responses.response:
        >>>         print(response)
        """
        return [r.response for r in self._responses]

    @property
    def correct(self):
        r"""A list of all correct (0, 1, or None) values.

        Examples
        --------
        >>> for correct in responses.correct:
        >>>         print(correct)
        """
        return [r.correct for r in self._responses]

    @property
    def response_time(self):
        r"""A list of all response times (float or None).

        Examples
        --------
        >>> for rt in responses.response_time:
        >>>         print(rt)
        """
        return [r.response_time for r in self._responses]

    @property
    def item(self):
        r"""A list of all item names (str or None) associated with each
        response.

        Examples
        --------
        >>> for item in responses.item:
        >>>         print(item)
        """
        return [r.item for r in self._responses]

    @property
    def feedback(self):
        r"""A list of the feedback status (True or False) associated with each
        response.

        Examples
        --------
        >>> for feedback in responses.feedback:
        >>>         print(feedback)
        """
        return [r.feedback for r in self._responses]

    @property
    def var(self):

        return self._experiment.var

    def add(self, response=None, correct=None, response_time=None, item=None,
            feedback=True):
        r"""Adds a response.

        Parameters
        ----------
        response    The response value, for example, 'space' for the spacebar, 0 for
            joystick button 0, etc.
        correct : bool, int, None, optional
            The correctness of the response.
        response_time : float, int, None, optional
            The response_time.
        item : str, None, optional
            The item that collected the response.
        feedback : bool, optional
            Indicates whether the response should be included in feedback on
            accuracy and average response time.

        Examples
        --------
        >>> responses.add(response_time=500, correct=1, response='left')
        """
        r = ResponseInfo(self, response=response, correct=correct,
                         response_time=response_time, item=item, feedback=feedback)
        if correct is None:
            correct = u'undefined'
        else:
            correct = r.correct
        self._responses.insert(0, r)
        self.var.response = self._experiment.syntax.auto_type(
            self._experiment.syntax.sanitize(r.response),
            allow_none=True,
            allow_bool=True)
        self.var.response_time = r.response_time
        self.var.correct = correct
        if item is not None:
            self.var.set(u'response_%s' % item,
                         self._experiment.syntax.sanitize(r.response))
            self.var.set(u'response_time_%s' % item, r.response_time)
            self.var.set(u'correct_%s' % item, correct)
        self.var.acc = self.var.accuracy = self.acc
        self.var.avg_rt = self.var.average_response_time = self.avg_rt
        # Old variables, mostly for backwards compatibility
        rs = self._select(feedback=True)
        self.var.total_response_time = sum(
            rs._selectnot(response_time=None).response_time)
        self.var.total_responses = len(rs)
        self.var.total_correct = len(rs._select(correct=1))

    def clear(self):
        r"""Clears all responses.

        Examples
        --------
        >>> responses.clear()
        """
        self._responses = []

    def reset_feedback(self):
        r"""Sets the feedback status of all responses to False, so that only
        new responses will be included in feedback.

        Examples
        --------
        >>> responses.reset_feedback()
        """
        for r in self._responses:
            r.feedback = False

    def _select(self, **kwdict):

        rs = ResponseStore(self._experiment)
        for r in self._responses:
            if r.match(**kwdict):
                rs._responses.append(r)
        return rs

    def _selectnot(self, **kwdict):

        rs = ResponseStore(self._experiment)
        for r in self._responses:
            if r.matchnot(**kwdict):
                rs._responses.append(r)
        return rs

    def __len__(self):

        return len(self._responses)

    def __getitem__(self, key):

        if isinstance(key, slice):
            rs = ResponseStore(self._experiment)
            rs._responses = self._responses[key]
            return rs
        if isinstance(key, int):
            return self._responses[key]
        raise TypeError(u'A key for responses should be either slice or int')

    def __str__(self):

        s = u'%d responses (last response is shown first):\n' % len(self)
        for i, r in enumerate(self):
            s += '%d: %s\n' % (i, r)
        return s

    def __iter__(self):

        for r in self._responses:
            yield r


# Alias for backwards compatibility
response_info = ResponseInfo
response_store = ResponseStore



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from qtpy import QtWidgets
from libqtopensesame.misc.base_subcomponent import BaseSubcomponent
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'item_combobox', category=u'core')


class ItemComboBox(QtWidgets.QComboBox, BaseSubcomponent):

    r"""A combobox to select existing items."""
    def __init__(self, parent=None, filter_fnc=None):

        QtWidgets.QComboBox.__init__(self, parent)
        self.setup(parent)
        self.filter_fnc = filter_fnc
        self.refresh()

    @property
    def filter_fnc(self):

        return self._filter_fnc

    @filter_fnc.setter
    def filter_fnc(self, filter_fnc):

        if filter_fnc is None:
            self._filter_fnc = lambda item: True
        else:
            self._filter_fnc = filter_fnc

    @property
    def selected_item(self):

        return self.currentText()

    def currentText(self):

        if self.currentIndex() == 0:
            return u''
        return QtWidgets.QComboBox.currentText(self)

    def findText(self, item):

        if not item:
            return 0
        return QtWidgets.QComboBox.findText(self, safe_decode(item))

    @property
    def items(self):

        return filter(self._filter_fnc, self.item_store)

    def select(self, item_name):

        i = self.findText(item_name)
        if i < 0:
            self.setCurrentIndex(0)
            return False
        self.setCurrentIndex(i)
        return True

    def refresh(self):

        prev_selected = self.selected_item
        while self.count():
            self.removeItem(0)
        self.addItem(_(u'No item selected'))
        self.setItemIcon(0, self.theme.qicon(u'go-down'))
        for i, item_name in enumerate(self.items):
            self.addItem(item_name)
            self.setItemIcon(i+1,
                             self.theme.qicon(self.item_store[item_name].item_icon()))
        self.select(prev_selected)


# Alias for backwards compatibility
item_combobox = ItemComboBox



# -*- coding:utf-8 -*-

"""
This file is part of openexp.

openexp is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

openexp is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with openexp.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libqtopensesame.misc.config import cfg
from libqtopensesame.sketchpad_elements._base_rect_ellipse import \
    BaseRectEllipse
from libopensesame.sketchpad_elements import Rect as RectRuntime


class Rect(BaseRectEllipse, RectRuntime):

    r"""A rect element.
    See base_element for docstrings and function
    descriptions.
    """
    @classmethod
    def mouse_press(cls, sketchpad, pos):

        if cls.pos_start is not None:
            return
        cls.pos_start = pos
        xc = sketchpad.canvas.xcenter()
        yc = sketchpad.canvas.ycenter()
        cls.preview = sketchpad.canvas.rect(pos[0]+xc, pos[1]+yc, 0, 0,
                                            color=cfg.sketchpad_preview_color,
                                            penwidth=cfg.sketchpad_preview_penwidth)


# Alias for backwards compatibility
rect = Rect



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *


class QtStructureItem:

    r"""A base class for items that control the structure of the experiment,
    i.e. those that require rebuilding of the item tree when they are changed.
    """
    def __init__(self):
        r"""Constructor."""
        self._children = None

    def update(self):
        """See qtitem."""
        super().update()
        self.experiment.build_item_tree()

    def apply_script_changes(self):
        """See qtitem."""
        super().apply_script_changes()
        self.experiment.build_item_tree()

    @staticmethod
    def clears_children_cache(fnc):
        r"""A decorator for functions that change the structure of the
        experiment, and thus need a clearing of the children cache.
        """
        def inner(self, *args, **kwargs):

            self.experiment.items.clear_cache()
            return fnc(self, *args, **kwargs)

        return inner

    @staticmethod
    def cached_children(fnc):
        r"""A decorator for the children function, which is cached to speed up
        performance.
        """
        def inner(self, *args, **kwargs):

            if self._children is not None:
                return self._children
            return fnc(self, *args, **kwargs)

        return inner


# Alias for backwards compatibility
qtstructure_item = QtStructureItem



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.item import Item
import fnmatch
from libopensesame.oslogging import oslogger
from libopensesame.exceptions import InvalidOpenSesameScript


class Logger(Item):
    """The logger item logs experimental data (i.e. variables)."""
    
    description = 'Logs experimental data'
    is_oneshot_coroutine = True

    def reset(self):
        self.logvars = []
        self._logvars = None
        self.var.auto_log = 'yes'
        self.exclude_patterns = []
        self._exclude_vars = []
        
    def _is_excluded(self, var):
        if var in self._logvars:
            return False
        if var in self._exclude_vars:
            return True
        for pattern in self.exclude_patterns:
            if fnmatch.fnmatch(var, pattern):
                self._exclude_vars.append(var)
                return True
        return False

    def run(self):
        self.set_item_onset()
        if self._logvars is None:
            if self.var.auto_log == 'yes':
                self._logvars = self.experiment.log.all_vars()
            else:
                self._logvars = []
            for var in self.logvars:
                if var not in self._logvars and not self._is_excluded(var):
                    self._logvars.append(var)
            self._logvars.sort()
        # If we are automatically logging all variables, emit a warning when
        # a new variable was created after the first logger call, because it
        # will not be logged in that case.
        elif self.var.auto_log == 'yes':
            for var in self.experiment.log.all_vars():
                if var not in self._logvars and not self._is_excluded(var):
                    oslogger.warning(
                        f'the variable {var} was created after the first '
                        f'logger call and will therefore not be logged.')
        self.experiment.log.write_vars(self._logvars)

    def coroutine(self, coroutines):
        yield
        self.run()

    def from_string(self, string):
        self.var.clear()
        self.comments = []
        self.reset()
        if string is None:
            return
        for line in string.split('\n'):
            self.parse_variable(line)
            cmd, arglist, kwdict = self.experiment.syntax.parse_cmd(line)
            if cmd in ('log', 'exclude') and len(arglist) > 0:
                if cmd == 'log':
                    for var in arglist:
                        if not self.experiment.syntax.valid_var_name(
                                safe_decode(var)):
                            oslogger.error(
                                f'{var} is not a valid variable name')
                    self.logvars += arglist
                else:
                    self.exclude_patterns += arglist

    def to_string(self):
        s = super().to_string('logger')
        for var in self.logvars:
            s += '\t' + self.experiment.syntax.create_cmd(
                'log', [var]) + '\n'
        for var in self.exclude_patterns:
            s += '\t' + self.experiment.syntax.create_cmd(
                'exclude', [var]) + '\n'
        return s


# Alias for backwards compatibility
logger = Logger



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.exceptions import InvalidFormScript, OSException, \
    InvalidValue
from libopensesame.item import Item
from libopensesame import widgets


class FormBase(Item):

    initial_view = 'script'

    def reset(self):
        self.var.cols = '2;2'
        self.var.rows = '2;2'
        self.var.spacing = 10
        self.var._theme = 'gray'
        self.var.only_render = 'no'
        self.var.timeout = 'infinite'
        self.var.margins = '50;50;50;50'
        self._widgets = []
        self._variables = []

    def set_validator(self):
        self.validator = FormBase

    def parse_line(self, line):
        cmd, arglist, kwdict = self.syntax.parse_cmd(line)
        if cmd != 'widget':
            return
        if len(arglist) != 5:
            raise InvalidFormScript(f'Invalid widget specification: {line}')
        self._widgets.append((arglist, kwdict))
        if 'var' in kwdict:
            self._variables.append(kwdict['var'])

    def to_string(self):
        s = super().to_string(self.item_type)
        for arglist, kwdict in self._widgets:
            s += '\t%s\n' % self.syntax.create_cmd('widget', arglist, kwdict)
        s += '\n'
        return s

    def run(self):
        self.set_item_onset()
        if self.var.only_render == 'yes':
            self._form.render()
        else:
            self._form._exec(focus_widget=self.focus_widget)

    def prepare(self):
        super().prepare()
        # Prepare the form
        try:
            cols = [float(i) for i in str(self.var.cols).split(';')]
            rows = [float(i) for i in str(self.var.rows).split(';')]
            margins = [float(i) for i in str(self.var.margins).split(';')]
        except Exception as e:
            raise InvalidValue('cols, rows, and margins should be numeric '
                               'values separated by a semi-colon')
        if self.var.timeout == 'infinite':
            timeout = None
        else:
            timeout = self.var.timeout
        self._form = widgets.Form(self.experiment, cols=cols, rows=rows,
                                  margins=margins, spacing=self.var.spacing,
                                  theme=self.var._theme,
                                  item=self, timeout=timeout,
                                  clicks=self.var.form_clicks == 'yes')

        self.focus_widget = None
        for arglist, orig_kwdict in self._widgets:
            kwdict = orig_kwdict.copy()
            # Evaluate all values
            arglist = [self.syntax.eval_text(arg, include_local=True)
                       for arg in arglist]
            for key, val in kwdict.items():
                kwdict[key] = self.syntax.eval_text(val, include_local=True)
            # Translate paths into full file names
            if 'path' in kwdict:
                kwdict['path'] = self.experiment.pool[kwdict['path']]
            # Process focus keyword
            focus = False
            if 'focus' in kwdict:
                if kwdict['focus'] == 'yes':
                    focus = True
                del kwdict['focus']
            # Parse arguments
            _type = arglist[4]
            try:
                col = int(arglist[0])
                row = int(arglist[1])
                colspan = int(arglist[2])
                rowspan = int(arglist[3])
            except (ValueError, TypeError):
                raise InvalidValue('In a form widget col, row, colspan, and '
                                   'rowspan should be integer')
            # Create the widget and add it to the form
            try:
                cls = getattr(widgets, _type)
            except AttributeError:
                raise InvalidFormScript(f'{_type} is not a valid form widget')
            _w = cls(self._form, **kwdict)
            self._form.set_widget(_w, (col, row), colspan=colspan,
                                  rowspan=rowspan)
            # Add as focus widget
            if focus:
                if self.focus_widget is not None:
                    raise InvalidFormScript(
                        'You can only specify one focus widget')
                self.focus_widget = _w

    def var_info(self):
        return super().var_info() + \
            [(var, '[Response variable]') for var in self._variables]



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from openexp._canvas.canvas import Canvas
from openexp._mouse.mouse import Mouse


class Coordinates:
    r"""A base class for classes that need to perform coordinate conversions.
    """
    
    def __init__(self):
        r"""Constructor."""
        self._width = self.experiment.var.width
        self._height = self.experiment.var.height
        self._xcenter = self._width/2
        self._ycenter = self._height/2
        self._bottom = self._ycenter
        self._top = -self._ycenter
        self._left = -self._xcenter
        self._right = self._xcenter
        self._mouse_dev = isinstance(self, Mouse)
        self._canvas_dev = isinstance(self, Canvas)
        if not self._mouse_dev and not self._canvas_dev:
            raise TypeError('coordinates class should be coparent with canvas '
                            'or mouse class')

    def none_to_center(self, x, y):
        r"""Interprets None coordinates as the display center.

        Parameters
        ----------
        x : int, float, NoneType
            An X coordinate.
        y : int, float, NoneType
            A Y coordinate.

        Returns
        -------
        tuple
            An (x, y) coordinate tuple.
        """
        if x is None:
            x = 0
        if y is None:
            y = 0
        return x, y

    def to_xy(self, x, y=None):
        r"""Converts coordinates from the OpenSesame reference frame to the
        back-end specific reference frame. `None` values are taken as the
        display center.

        Parameters
        ----------
        x : float, int, NoneType, tuple
            An x coordinate, or an (x,y) tuple.
        y : float, int, NoneType, optional
            A y coordinate. Only applicable if x was not a tuple.

        Returns
        -------
        tuple
            An (x, y) coordinate tuple in the back-end specific reference
            frame.
        """
        raise NotImplementedError()

    def from_xy(self, x, y=None):
        r"""Converts coordinates from the back-end specific reference frame to
        the OpenSesame reference frame.

        Parameters
        ----------
        x : float, int, tuple
            An x coordinate, or an (x,y) tuple.
        y : float, int, NoneType, optional
            A y coordinate. Only applicable if x was not a tuple.

        Returns
        -------
        tuple
            An (x, y) coordinate tuple in the OpenSesame reference frame.
        """
        raise NotImplementedError()


# Non PEP-8 alias for backwards compatibility
coordinates = Coordinates



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
import os
import sys
from openexp import resources
from libopensesame import misc, metadata
from libqtopensesame.extensions import BaseExtension
from libqtopensesame.misc.translate import translation_context
from libqtopensesame.misc import template_info
_ = translation_context(u'get_started', category=u'extension')


class GetStarted(BaseExtension):
    r"""Shows the get-started tab and opens an experiment on startup, if one
    was passed on the command line.
    """
    
    def activate(self):
        # Initialize templates
        templates = []
        for i, (path, desc) in enumerate(template_info.templates):
            try:
                path = resources[path]
            except:
                continue
            if not i:
                cls = u'important-button'
            else:
                cls = u'button'
            # We need to use forward slashes, also on Windows, otherwise it
            # leads to an invalid URL when prefixed with opensesame://
            path = os.path.abspath(path).replace(u'\\', u'/')
            md = u'<a href="opensesame://%s" class="%s">%s</a><br />' \
                % (path, cls, desc)
            templates.append(md)
        # Initialize recent experiments
        if not self.main_window.recent_files:
            recent = []
        else:
            recent = [_(u'Continue with a recent experiment:')+u'<br />']
            for i, path in enumerate(self.main_window.recent_files):
                cls = u'important-button' if not i else u'button'
                md = u'<a href="opensesame://event.open_recent_%d" class="%s">%s</a><br />' % \
                    (i, cls, self._unambiguous_path(path))
                recent.append(md)
        # Create markdown
        with safe_open(self.ext_resource(u'get_started.md')) as fd:
            md = fd.read()
        md = md % {
            u'version': metadata.__version__,
            u'codename': metadata.codename,
            u'templates': u'  \n'.join(templates),
            u'recent_experiments': u'  \n'.join(recent)
        }
        self.tabwidget.open_markdown(md, title=_(u'Get started!'),
                                     icon=u'document-new')

    def _unambiguous_path(self, path):
        r"""If the path basename is unique among the resent experiments, this
        is used. Otherwise, the full path is used.

        Parameters
        ----------
        path
            The path to shorten unambiguously.

        Returns
        -------
        The unambiguously shortened path.
        """
        basename = os.path.basename(path)
        basenames = \
            [os.path.basename(_path)
             for _path in self.main_window.recent_files]
        return path if basenames.count(basename) > 1 else basename

    def event_open_recent_0(self):
        self.main_window.open_file(path=self.main_window.recent_files[0])

    def event_open_recent_1(self):
        self.main_window.open_file(path=self.main_window.recent_files[1])

    def event_open_recent_2(self):
        self.main_window.open_file(path=self.main_window.recent_files[2])

    def event_open_recent_3(self):
        self.main_window.open_file(path=self.main_window.recent_files[3])

    def event_open_recent_4(self):
        self.main_window.open_file(path=self.main_window.recent_files[4])

    @BaseExtension.as_thread(wait=500)
    def event_startup(self):
        # Open an experiment if it has been specified as a command line
        # argument and suppress the new wizard in that case.
        if len(sys.argv) >= 2 and os.path.isfile(sys.argv[1]):
            path = safe_decode(sys.argv[1], enc=sys.getfilesystemencoding(),
                               errors=u'ignore')
            self.main_window.open_file(path=path)
            return
        self.activate()



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from openexp._canvas._image.image import Image
from openexp._canvas._element.psycho import PsychoElement, RotatingElement
from psychopy import visual


class Psycho(RotatingElement, PsychoElement, Image):

    def prepare(self):

        self._stim = visual.ImageStim(
            win=self.win,
            image=safe_decode(self.fname)
        )
        if self.rotation is not None and self.rotation != 0:
            self._stim.ori = self.rotation
        if self.scale is not None:
            w, h = self._stim.size
            w *= self.scale
            h *= self.scale
            self._stim.size = w, h
        x, y = self.to_xy(self.x, self.y)
        if not self.center:
            size = self._stim.size
            x += size[0] / 2
            y -= size[1] / 2
        self._stim.pos = x, y



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from qtpy import QtWidgets, QtCore
from libqtopensesame.misc.base_subcomponent import BaseSubcomponent
from libqtopensesame.widgets import pool_widget
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'pool_select', category=u'core')


class PoolSelect(QtWidgets.QWidget, BaseSubcomponent):
    """A widget that implements a file-pool selector. Partly emulates the
    QLineEdit API.
    """
    editingFinished = QtCore.Signal()
    textEdited = QtCore.Signal()

    def __init__(self, parent=None):
        QtWidgets.QComboBox.__init__(self, parent)
        self.setup(parent)
        self.edit = QtWidgets.QLineEdit()
        self.edit.editingFinished.connect(self.editingFinished.emit)
        self.edit.textEdited.connect(self.textEdited.emit)
        self.edit.setMinimumWidth(200)
        self.button = QtWidgets.QPushButton(self.theme.qicon('browse'),
                                            _('Browse'))
        self.button.setIconSize(QtCore.QSize(16, 16))
        self.button.clicked.connect(self.browse)
        self.hbox = QtWidgets.QHBoxLayout(self)
        self.hbox.addWidget(self.edit)
        self.hbox.addWidget(self.button)
        self.hbox.setContentsMargins(0, 0, 0, 0)
        self.hbox.setSpacing(6)
        self.setLayout(self.hbox)

    def text(self):
        return self.edit.text()

    def setText(self, value):
        return self.edit.setText(value)

    def browse(self):
        s = pool_widget.select_from_pool(self.main_window)
        if not s:
            return
        self.edit.setText(s)
        self.editingFinished.emit()


# Alias for backwards compatibility
pool_select = PoolSelect



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame import type_check
from libopensesame.exceptions import InvalidFormGeometry
from openexp.canvas import canvas
from openexp.mouse import mouse
from openexp.keyboard import keyboard
from openexp._keyboard.keybabel import KeyBabel
from libopensesame.widgets.widget_factory import WidgetFactory


class Form:

    r"""The `Form` is a container for widgets, such as labels, etc. If you use
    the FORM_BASE plug-in in combination with OpenSesame script, you do not
    need to explicitly create a `Form` object. However, if you use Python
    inline code, you do.

    __Example__:

    ~~~ .python
    form = Form()
    label =
    Label(text='label)
    form.set_widget(label, (0,0))
    form._exec()
    ~~~

    [TOC]
    """
    def __init__(self, experiment, cols=2, rows=2, spacing=10,
                 margins=(100, 100, 100, 100), theme=u'gray', item=None, timeout=None,
                 clicks=False, validator=None):
        r"""Constructor to create a new `Form` object. You do not generally
        call this constructor directly, but use the `Form()` factory
        function,
        which is described here: [/python/common/]().

        Parameters
        ----------
        experiment : experiment
            An OpenSesame experiment.
        cols : int, list, optional
            The number of columns (as int) or a list that specifies the number
            and relative size of the columns. For example, `[1,2,1]` will
            create 3 columns where the middle one is twice as large as the
            outer ones.
        rows : int, list, optional
            Analogous to `cols`, but for the rows.
        spacing : int, optional
            The amount of empty space between widgets (in pixels).
        margins : list, optional
            The amount of empty space around the form. This is specified as a
            list, like so [top-margin, right-margin, bottom-margin, left-
            margin].
        theme : str, unicode, optional
            The theme for the widgets.
        item : item, NoneType, optional
            The item of which the form is part.
        timeout : int, float, NoneType, optional
            A timeout value in milliseconds, or None if no timeout exists.
        clicks : bool, optional
            If enabled, an auditory click is played on user interactions. This
            can help to make interactions feel smoother if there is some visual
            lag.
        validator : FunctionType, NoneType, optional
            A function that takes no arguments and returns True if the form is
            successfully validated, and False if not.
        """
        # Normalize the column and row sizes so that they add up to 1
        if isinstance(cols, int):
            self.cols = [1./cols]*cols
        else:
            cols = type_check.float_list(cols, u'form columns', min_len=1)
            self.cols = [float(c)/sum(cols) for c in cols]
        if isinstance(rows, int):
            self.rows = [1./rows]*rows
        else:
            rows = type_check.float_list(rows, u'form rows', min_len=1)
            self.rows = [float(r)/sum(rows) for r in rows]
        self.experiment = experiment
        self.item = item if item is not None else experiment
        self.timeout = timeout
        self.width = experiment.var.width
        self.height = experiment.var.height
        self.spacing = spacing
        self.clicks = clicks
        self.margins = type_check.float_list(margins, u'form margins',
                                             min_len=4, max_len=4)
        n_cells = len(self.cols)*len(self.rows)
        self.widgets = [None]*n_cells
        self.span = [(1, 1)]*n_cells
        self._validator = (lambda: True) if validator is None else validator
        self.canvas = canvas(
            self.experiment,
            color=self.item.var.foreground,
            background_color=self.item.var.background)
        # Maps ids of WidgetFactory objects to corresponding Widgets.
        self._widget_factory_map = {}
        # Dynamically load the theme object
        theme_mod = __import__(
            u'libopensesame.widgets.themes.%s' % theme, fromlist=[u'dummy'])
        theme_cls = getattr(theme_mod, theme)
        self.theme_engine = theme_cls(self)

    def __len__(self):
        r"""Implements the `len()` syntax.

        Returns
        -------
        int
            The number of widgets in the form.
        """
        return sum([w is not None for w in self.widgets])

    def render(self):
        r"""Shows the form canvas without any user interaction."""
        self.canvas.show()

    def _exec(self, focus_widget=None):
        r"""Executes the form.

        Parameters
        ----------
        focus_widget : widget, NoneType, optional
            A widget that is in the form and should receive a virtual mouse
            click when the form is opened. This allows you to activate a
            text_input right away, for example, so that the user doesn't have
            to click on it anymore.

        Returns
        -------
        """
        if focus_widget is not None:
            focus_widget = self._widget_instance(focus_widget)
            focus_widget.focus = True
        if len(self) == 0:
            raise InvalidFormGeometry('The form contains no widgets')
        ms = mouse(self.experiment, buttonlist=[1, 2, 3], timeout=0)
        ms.show_cursor()
        kb = keyboard(self.experiment, timeout=0)
        kb.show_virtual_keyboard()
        babel = KeyBabel(kb)
        coroutines = {w: w.coroutine() for w in self.widgets if w is not None}
        for coroutine in coroutines.values():
            coroutine.send(None)
        self.canvas.show()
        self.start_time = None
        mousedown = False
        while True:
            if self.timed_out():
                resp = None
                break
            msg = None
            # Handle mouse clicks, including waiting until the mouse is released
            # after a mouse click.
            if mousedown:
                while any(ms.get_pressed()):
                    pass
                mousedown = False
            button, xy, timestamp = ms.get_click(visible=True)
            if button is not None:
                mousedown = True
                # Switch the focus to the newly clicked widget (if any)
                widget = self.xy_to_widget(xy)
                if widget is None:
                    continue
                if focus_widget is not None:
                    focus_widget.focus = False
                widget.focus = True
                focus_widget = widget
                msg = {
                    u'type': u'click',
                    u'pos': xy,
                    u'button': button,
                    u'timestamp': timestamp
                }
            # Handle key presses clicks
            elif focus_widget is not None:
                key, timestamp = kb.get_key()
                modifiers = kb.get_mods()
                if key is not None:
                    msg = {
                        u'type': u'key',
                        u'key': babel.standard_name(
                            key,
                            shift=u'shift' in modifiers
                        ),
                        u'timestamp': timestamp
                    }
            # Send message (if any)
            if msg is None:
                continue
            resp = coroutines[focus_widget].send(msg)
            self.canvas.show()
            if resp is not None and self._validator():
                break
        kb.show_virtual_keyboard(False)
        ms.show_cursor(False)
        for coroutine in coroutines.values():
            try:
                coroutine.send({u'type': u'stop'})
            except StopIteration:
                pass
        self.experiment.var.form_response = resp
        return resp

    def timed_out(self):
        """
        visible: False

        returns:
                desc:	True if a timeout occurred, False otherwise.
                type:	bool
        """
        if self.timeout is None:
            return False
        if self.start_time is None:
            self.start_time = self.item.clock.time()
            return False
        return self.item.clock.time()-self.start_time >= self.timeout

    def cell_index(self, pos):
        r"""Converts a position to a cell index. A cell index corresponds to
        the number of the cell in the form, from left-to-right, top-to-bottom.

        Parameters
        ----------
        pos : int, tuple
            A position in the form, which can be an index (int) or a (column,
            row) tuple.

        Returns
        -------
        int
            A cell index.
        """
        if type(pos) == int:
            return pos
        if type(pos) in (tuple, list) and len(pos) == 2:
            return pos[1]*len(self.cols)+pos[0]
        raise InvalidFormGeometry(f'{pos} is an invalid position in the form')

    def validate_geometry(self):
        for index1 in range(len(self.widgets)):
            if self.widgets[index1] is None:
                continue
            l = self.get_cell(index1)
            colspan, rowspan = self.span[index1]
            for col in range(l[0], l[0]+colspan):
                for row in range(l[1], l[1]+rowspan):
                    index2 = self.cell_index((col, row))
                    if index1 == index2:
                        continue
                    if len(self.widgets) <= index2:
                        raise InvalidFormGeometry(
                            f'The widget at position ({l[0]}, {l[1]}) falls '
                            f'outside of your form')
                    if self.widgets[index2] is not None:
                        raise InvalidFormGeometry(
                            f'The widget at position ({l[0]}, {l[1]}) '
                            f'overlaps with another widget')

    def get_cell(self, index):
        """
        Returns the position of a widget

        Arguments:
        index -- the index of the widget

        Returns:
        A (column, row, column_span, row_span) tuple
        """
        index = self.cell_index(index)
        col = index % len(self.cols)
        row = index // len(self.cols)
        colspan, rowspan = self.span[index]
        return col, row, colspan, rowspan

    def get_rect(self, index):
        """
        Returns the boundary area for a given cell

        Arguments:
        index -- a cell index

        Returns:
        A (left, top, width, height) tuple
        """
        col = index % len(self.cols)
        row = index // len(self.cols)
        colspan, rowspan = self.span[index]
        effective_width = self.width-self.margins[1]-self.margins[3]
        effective_height = self.height-self.margins[0]-self.margins[2]
        x1 = effective_width*sum(self.cols[:col])+self.spacing
        y1 = effective_height*sum(self.rows[:row])+self.spacing
        x2 = effective_width*sum(self.cols[:col+colspan])-self.spacing
        y2 = effective_height*sum(self.rows[:row+rowspan])-self.spacing
        w = x2-x1
        h = y2-y1
        if w <= 0 or h <= 0:
            raise InvalidFormGeometry('There is not enough space to show '
                                       'some form widgets. Please modify the '
                                       'form geometry!')
        x = x1+self.margins[3]
        y = y1+self.margins[0]
        x -= self.width/2
        y -= self.height/2
        return x, y, w, h

    def set_widget(self, widget, pos, colspan=1, rowspan=1):
        r"""Adds a widget to the form.

        Parameters
        ----------
        widget : widget
            The widget to add.
        pos : int, tuple
            The position to add the widget, which can be an index or a (column,
            row) tuple.
        colspan : int, optional
            The number of columns that the widget should span.
        rowspan : int, optional
            The number of rows that the widget should span.
        """
        index = self.cell_index(pos)
        if index >= len(self.widgets):
            raise InvalidFormGeometry(
                f'Widget position ({pos}) is outside of the form')
        if not isinstance(colspan, int) or colspan < 1 or colspan > len(self.cols):
            raise InvalidFormGeometry(
                f'Column span {colspan} is invalid (i.e. too large, too '
                f'small, or not a number)')
        if not isinstance(rowspan, int) or rowspan < 1 or rowspan > len(self.rows):
            raise InvalidFormGeometry(
                f'Row span {rowspan} is invalid (i.e. too large, too '
                f'small, or not a number)')
        widget = self._widget_instance(widget)
        self.widgets[index] = widget
        self.span[index] = colspan, rowspan
        widget.set_rect(self.get_rect(index))

    def xy_to_index(self, xy):
        r"""Converts a coordinate in pixels to a cell index. This allows you to
        determine on which widget a user has clicked.

        Parameters
        ----------
        xy : tuple
            An (x,y) coordinates tuple.

        Returns
        -------
        int
            A cell index.
        """
        for index in range(len(self.widgets)):
            x, y, w, h = self.get_rect(index)
            if x <= xy[0] and x+w >= xy[0] and y <= xy[1] and y+h >= xy[1]:
                return index
        return None

    def xy_to_widget(self, xy):

        index = self.xy_to_index(xy)
        if index is None:
            return None
        return self.widgets[index]
        
    def _widget_instance(self, widget):
        """Takes a Widget or WidgetFactory and returns a Widget. If a 
        WidgetFactory has already been instantiated before, the Widget instance
        is retrieved and returned.
        """
        if isinstance(widget, WidgetFactory):
            factory_id = id(widget)
            if factory_id in self._widget_factory_map:
                return self._widget_factory_map[factory_id]
            widget = widget.construct(self)
            self._widget_factory_map[factory_id] = widget
        return widget


form = Form



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.oslogging import oslogger
import warnings


class Log:
    """The `log` object provides data logging. A `log` object is created
    automatically when the experiment starts.

    __Example__

    ~~~ .python
    # Write one line of text
    log.write('My custom log message')
    # Write all variables
    log.write_vars()
    ~~~

    [TOC]
    """
    def __init__(self, experiment, path):
        r"""Constructor to create a new `log` object. You do not generally call
        this constructor directly, because a `log` object is created
        automatically when the experiment is launched.

        Parameters
        ----------
        experiment : experiment
            The experiment object.
        """
        self.experiment = experiment
        self.experiment.var.logfile = path
        self._all_vars = None
        self.open(path)

    def __call__(self, msg):

        self.write(msg)

    def close(self):
        r"""Closes the current log.

        Examples
        --------
        >>> log.close()
        """
        pass

    def all_vars(self):
        """{nodoc}
        
        returns:
                A list of all variables that exist in the experiment.
        """
        all_vars = []
        for key in self.experiment.var.inspect().keys():
            try:
                key = self.experiment.syntax.eval_text(key)
            except Exception:
                oslogger.warning(f'cannot evaluate variable name {key}')
            else:
                all_vars.append(key)
        return all_vars

    def open(self, path):
        r"""Opens the current log. If a log was already open, it is closed
        automatically, and re-opened.

        Parameters
        ----------
        path : str, unicode
            The path to the current logfile. In most cases (unless) a custom
            log back-end is used, this will be a filename.

        Examples
        --------
        >>> # Open a new log
        >>> log.open('/path/to/new/logfile.csv')
        """
        pass

    def write(self, msg, newline=True):
        r"""Write one message to the log.

        Parameters
        ----------
        msg : str, unicode
            A text message. When using Python 2, this should be either
            `unicode` or a utf-8-encoded `str`. When using Python 3, this
            should be either `str` or a utf-8-encoded `bytes`.
        newline : bool, optional
            Indicates whether a newline should be written after the message.

        Examples
        --------
        >>> # Write a single string of text
        >>> log.write(f'time = {clock.time()}')
        """
        pass

    def write_vars(self, var_list=None):
        r"""Writes variables to the log.

        Parameters
        ----------
        var_list : list, NoneType, optional
            A list of variable names to write, or None to write all variables
            that exist in the experiment.

        Examples
        --------
        >>> # Write all variables to the logfile
        >>> log.write_vars()
        """
        pass


# Non PEP-8 alias for backwards compatibility
log = Log



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from collections import OrderedDict
from libopensesame.oslogging import oslogger
from libopensesame.exceptions import OSException, InvalidValue
from openexp.backend import configurable
from openexp._canvas.canvas import Canvas
from openexp._coordinates.psycho import Psycho as PsychoCoordinates
from openexp.color import Color
from openexp.canvas_elements import Rect
from psychopy import visual, event, core, logging

# Store the experiment as a singleton, to be used in the _time() function
_experiment = None
# Store the old display gamma value
_old_gamma = None


class Psycho(Canvas, PsychoCoordinates):

    r"""This is a canvas backend built on top of PsychoPy (with Pyglet). For
    function specifications and docstrings, see `openexp._canvas.canvas`.
    """
    # The settings variable is used by the GUI to provide a list of back-end
    # settings
    settings = {
        u'psychopy_waitblanking': {
            u'name': u'Wait for blanking',
            u'description': u'Block until the display has been shown',
            u'default': u'yes'
        },
        u'psychopy_monitor': {
            u'name': u'Monitor',
            u'description': u'Virtual monitor',
            u'default': u'testMonitor'
        },
        u"psychopy_screen": {
            u"name": u"Screen",
            u"description": u"The physical screen that is used",
            u"default": 0,
        },
        u"psychopy_gamma": {
            u"name": u"Gamma",
            u"description": u"Display gamma value display",
            u"default": u"unchanged",
        },
        u'psychopy_suppress_warnings': {
            u'name': u'Suppress warnings',
            u'description': u'Set PsychoPy logging level to "critical"',
            u'default': u'yes',
        }
    }

    def __init__(self, experiment, auto_prepare=True, **style_args):

        Canvas.__init__(self, experiment, auto_prepare=auto_prepare,
                        **style_args)
        PsychoCoordinates.__init__(self)
        self.clear()

    @configurable
    def clear(self):

        self._elements = OrderedDict()
        self._set_background()

    def set_config(self, **cfg):

        Canvas.set_config(self, **cfg)
        if u'background_color' in cfg and hasattr(self, u'_elements'):
            self._set_background()

    def lower_to_bottom(self, element):

        Canvas.lower_to_bottom(self, element)
        Canvas.lower_to_bottom(self, u'__background__')

    def show(self):

        t0 = self.experiment.clock.time()
        for e in self._elements.values():
            e.show()
        self.experiment.window.flip(clearBuffer=True)
        t1 = self.experiment.clock.time()
        if t1 - t0 > self.MAX_SHOW_DT:
            oslogger.warning('Canvas.show() took {0} ms'.format(t1 - t0))
        return t1

    def _set_background(self):

        if u'__background__' in self:
            del self['__background__']
        self['__background__'] = Rect(
            self.left, self.top, self.width, self.height,
            color=self.background_color.colorspec, fill=True
        )
        Canvas.lower_to_bottom(self, u'__background__')

    def __iter__(self):

        for name, stim in self._elements.items():
            if name == u'__background__':
                continue
            yield name, stim

    def __len__(self):

        return len(self._elements)-1

    @staticmethod
    def init_display(experiment):

        if not py3:
            from json_tricks import utils
            utils.get_arg_names = _get_arg_names

        global _experiment, _old_gamma
        _experiment = experiment
        # Set the PsychoPy monitor, default to testMonitor
        monitor = experiment.var.get(u'psychopy_monitor', u'testMonitor')
        waitblanking = experiment.var.get(u'psychopy_waitblanking', u'yes',
                                          [u'yes', u'no']) == u'yes'
        screen = experiment.var.get(u'psychopy_screen', 0)
        # Print some information to the debug window
        oslogger.info(u'waitblanking = %s' % waitblanking)
        oslogger.info(u'monitor = %s' % monitor)
        oslogger.info(u'screen = %s' % screen)
        # Initialize the PsychoPy window and set various functions
        experiment.window = visual.Window(
            experiment.resolution(),
            screen=screen,
            waitBlanking=waitblanking,
            fullscr=experiment.var.fullscreen == u'yes',
            monitor=monitor,
            units=u'pix',
            color=Color(experiment, experiment.var.background).backend_color,
            winType=u'pyglet',
            allowStencil=True
        )
        event.Mouse(visible=False, win=experiment.window)
        experiment.window.winHandle.set_caption(
            u'OpenSesame (PsychoPy backend)')
        # Set Gamma value if specified
        gamma = experiment.var.get(u'psychopy_gamma', u'unchanged')
        if type(gamma) in (int, float) and gamma > 0:
            _old_gamma = experiment.window.gamma
            experiment.window.setGamma(gamma)
        elif gamma != u'unchanged':
            raise InvalidValue(
                u'Gamma should be a positive numeric value or "unchanged"')
        core.quit = _psychopy_clean_quit
        # Optionally change the logging level to avoid a lot of warnings in the
        # debug window
        if experiment.var.get(u'psychopy_suppress_warnings', u'yes'):
            logging.console.setLevel(logging.CRITICAL)

    @staticmethod
    def close_display(experiment):

        global _old_gamma
        # Restore display gamma if necessary
        if _old_gamma is not None:
            experiment.window.setGamma(_old_gamma)
        # This causes a (harmless) exception in some cases, so we catch it to
        # prevent confusion
        try:
            experiment.window.close()
        except:
            oslogger.error(
                u'An error occurred while closing the PsychoPy window.',
            )


def _get_arg_names(callable):
    """A hack to make json_tricks work on Python 2, after some unknown module
    has injected getfullargspec into inspect.
    """
    from functools import partial
    from logging import warn
    from sys import version_info
    from inspect import getargspec

    if type(callable) == partial and version_info[0] == 2:
        if not hasattr(get_arg_names, '__warned_partial_argspec'):
            get_arg_names.__warned_partial_argspec = True
            warn("'functools.partial' and 'inspect.getargspec' are not compatible in this Python version; "
                 "ignoring the 'partial' wrapper when inspecting arguments of {}, which can lead to problems".format(callable))
        return set(getargspec(callable.func).args)
    argspec = getargspec(callable)
    return set(argspec.args)


def _psychopy_clean_quit():
    r"""When PsychoPy encounters an error, it does a sys.exit() which is not
    what we want, because it closes OpenSesame altogether. Instead, we nicely
    inform the user that PsychoPy has signalled an error.
    """
    raise OSException(
        u'PsychoPy encountered an error and aborted the program. See the debug window for PsychoPy error messages.')


# Non PEP-8 alias for backwards compatibility
psycho = Psycho



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from openexp._coordinates.coordinates import Coordinates


class Xpyriment(Coordinates):
    r"""For function specifications and docstrings, see
    `openexp._coordinates.coordinates`.
    """
    
    def __init__(self):

        Coordinates.__init__(self)
        self._xwcenter = self.experiment.expyriment.screen.window_size[0]/2
        self._ywcenter = self.experiment.expyriment.screen.window_size[1]/2

    def to_xy(self, x, y=None):

        if isinstance(x, tuple):
            x, y = x
        x, y = self.none_to_center(x, y)
        if self._canvas_dev:
            return x, -y
        if self._mouse_dev:
            # The mouse is centered on the top-left, but we need to take into
            # account that the display is padded in fullscreen mode.
            return x + self._xwcenter, y + self._ywcenter

    def from_xy(self, x, y=None):

        if isinstance(x, tuple):
            x, y = x
        if not self._mouse_dev:
            raise TypeError('Only mouse supported')
        return x - self._xwcenter, y - self._ywcenter


# Non PEP-8 alias for backwards compatibility
xpyriment = Xpyriment



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
import time
from libopensesame.oslogging import oslogger
from libqtopensesame.misc import drag_and_drop
from libqtopensesame.misc.base_component import BaseComponent


class BaseSubcomponent(BaseComponent):

    r"""A base class for all components that require an experiment and theme
    property.
    """
    def name(self):
        """
        returns:
            desc:    The name of the extension, i.e. the extension class name.
            type:    unicode
        """
        return safe_decode(self.__class__.__name__, enc=u'utf-8')

    def fire(self, event, **kwdict):
        r"""Calls an event function, if it exists.

        Parameters
        ----------
        event : str, unicode
            The event name, which is handled by a function called `event_[event
            name]`.
        **kwdict : dict
            A keyword dictionary with event-specific keywords.
            
        Returns
        -------
        The return value of the event function, or None if no event function
        exists.
        """
        if hasattr(self, u'event_%s' % event):
            return getattr(self, u'event_%s' % event)(**kwdict)

    def _fire_and_time(self, event, **kwdict):
        r"""Similar to fire() but also keeps track of the duration of each
        event for benchmarking purposes. This function is used if the
        --performance-profile argument is passed.
        """
        if hasattr(self, u'event_%s' % event):
            t0 = time.time()
            getattr(self, u'event_%s' % event)(**kwdict)
            if event not in self._event_durations:
                self._event_durations[event] = []
            self._event_durations[event].append(time.time() - t0)

    def provide(self, provide, **kwdict):
        r"""Executing a providing function, if it exists.

        Parameters
        ----------
        provide : str, unicode
            The provide name, which is handled by a function called
            `provide_[event name]`.
        **kwdict : dict
            A keyword dictionary with provide-specific keywords.


        Returns
        -------
        The return value of the providing function.
        """
        if hasattr(self, u'provide_%s' % provide):
            return getattr(self, u'provide_%s' % provide)(**kwdict)

    def supported_events(self):
        r"""Gives the events that are supported by the extension. This is done
        by introspecting which `event_[event name]` functions exist.

        Returns
        -------
        list
            A list of supported events.
        """
        events = []
        for event in dir(self):
            if event.startswith(u'event_'):
                events.append(event[6:])
                oslogger.debug(
                    u'extension %s supports %s' % (self.name(), event)
                )
        return events

    def supported_provides(self):
        r"""Gives the provides that are supported by the extension. This is
        done by introspecting which `provide_[provide name]` functions exist.

        Returns
        -------
        list
            A list of supported provides.
        """
        provides = []
        for provide in dir(self):
            if provide.startswith(u'provide_'):
                provides.append(provide[8:])
                oslogger.debug(
                    u'extension %s provides %s' % (self.name(), provide)
                )
        return provides
    
    def report_exception(self, e):
        """Prints an exception to the console and shows a markdown tab.
        
        Parameters
        ----------
        e : Exception
        """
        self.console.write(e)
        self.tabwidget.open_markdown(e.markdown(), 'os-finished-error',
                                     e.title())

    # These properties allow for cleaner programming, without dot references.

    @property
    def experiment(self):
        return self.main_window.experiment

    @property
    def tabwidget(self):
        return self.main_window.ui.tabwidget

    @property
    def theme(self):
        return self.main_window.theme

    @property
    def notify(self):
        return self.main_window.notify

    @property
    def item_store(self):
        return self.experiment.items

    @property
    def extensions(self):
        return self.main_window.extension_manager._extensions

    @property
    def extension_manager(self):
        return self.main_window.extension_manager
    
    @property
    def unloaded_extension_manager(self):
        return self.main_window._unloaded_extension_manager
    
    @property
    def plugin_manager(self):
        return self.main_window.experiment._plugin_manager

    @property
    def overview_area(self):
        return self.main_window.ui.itemtree

    @property
    def console(self):
        return self.main_window.console

    @property
    def pool(self):
        return self.main_window.experiment.pool

    @property
    def locale(self):
        return self.main_window._locale

    @property
    def item_toolbar(self):
        return self.main_window.ui.toolbar_items

    @property
    def main_toolbar(self):
        return self.main_window.ui.toolbar_main


# Alias for backwards compatibility
base_subcomponent = BaseSubcomponent
